import collections
import types

try:
    import graphviz
    GRAPHVIZ_INSTALLED = True
except ImportError:
    GRAPHVIZ_INSTALLED = False

from .. import base

from . import func


__all__ = ['TransformerUnion']


class TransformerUnion(collections.UserDict, base.Transformer):
    """Packs multiple transformers into a single one.

    Calling ``transform_one`` will concatenate each transformer's output using a
    `collections.ChainMap`.

    Parameters:
        transformers (list): transformers to pack together.

    Example:

        ::

            >>> from pprint import pprint
            >>> import creme.compose
            >>> import creme.feature_extraction
            >>> import creme.stats

            >>> X = [
            ...     {'place': 'Taco Bell', 'revenue': 42},
            ...     {'place': 'Burger King', 'revenue': 16},
            ...     {'place': 'Burger King', 'revenue': 24},
            ...     {'place': 'Taco Bell', 'revenue': 58},
            ...     {'place': 'Burger King', 'revenue': 20},
            ...     {'place': 'Taco Bell', 'revenue': 50}
            ... ]

            >>> mean = creme.feature_extraction.Agg(
            ...     on='revenue',
            ...     by='place',
            ...     how=creme.stats.Mean()
            ... )
            >>> count = creme.feature_extraction.Agg(
            ...     on='revenue',
            ...     by='place',
            ...     how=creme.stats.Count()
            ... )
            >>> agg = creme.compose.TransformerUnion([mean])
            >>> agg += count

            >>> for x in X:
            ...     pprint(agg.fit_one(x).transform_one(x))
            {'revenue_count_by_place': 1, 'revenue_mean_by_place': 42.0}
            {'revenue_count_by_place': 1, 'revenue_mean_by_place': 16.0}
            {'revenue_count_by_place': 2, 'revenue_mean_by_place': 20.0}
            {'revenue_count_by_place': 2, 'revenue_mean_by_place': 50.0}
            {'revenue_count_by_place': 3, 'revenue_mean_by_place': 20.0}
            {'revenue_count_by_place': 3, 'revenue_mean_by_place': 50.0}

            >>> pprint(agg.transform_one({'place': 'Taco Bell'}))
            {'revenue_count_by_place': 3, 'revenue_mean_by_place': 50.0}

    """

    def __init__(self, transformers=None):
        super().__init__()
        if transformers is not None:
            for transformer in transformers:
                self += transformer

    @property
    def is_supervised(self):
        return any(transformer.is_supervised for transformer in self.values())

    def __str__(self):
        """Returns a human friendly representation of the pipeline."""
        return f' + '.join(map(str, self.keys()))

    def __repr__(self):
        return (
            'TransformerUnion (\n    ' +
            '    '.join(',\n'.join(map(repr, self.values())).splitlines(True)) +
            '\n)'
        )

    def add_step(self, other):
        """Adds a transformer while taking care of the input type."""

        # Infer a name if none is given
        if not isinstance(other, (list, tuple)):
            other = (str(other), other)
        name, transformer = other

        # If a function is given then wrap it in a FuncTransformer
        if isinstance(transformer, types.FunctionType):
            name = transformer.__name__
            transformer = func.FuncTransformer(transformer)

        # Prefer clarity to magic
        if name in self:
            raise KeyError(f'{name} already exists')

        # Store the transformer
        self[name] = transformer

        return self

    def __add__(self, other):
        return self.add_step(other)

    def fit_one(self, x, y=None):
        for transformer in self.values():
            transformer.fit_one(x, y)
        return self

    def transform_one(self, x):
        """Passes the data through each transformer and packs the results together."""
        return dict(collections.ChainMap(*(
            transformer.transform_one(x)
            for transformer in self.values()
        )))

    def draw(self):

        if not GRAPHVIZ_INSTALLED:
            raise ImportError('graphviz is not installed')

        g = graphviz.Digraph(engine='fdp')

        for part in self.values():
            if hasattr(part, 'draw'):
                g.subgraph(part.draw())
            else:
                g.node(str(part))
        return g



"""Loss functions."""
from libc cimport math


__all__ = [
    'Absolute',
    'Cauchy',
    'CrossEntropy',
    'Hinge',
    'EpsilonInsensitiveHinge',
    'Log',
    'Quantile',
    'Squared'
]


cdef double clamp_proba(double x):
    return math.fmax(math.fmin(x, 1 - 1e-15), 1e-15)


cdef class Loss:

    def __str__(self):
        return self.__class__.__name__


cdef class ClassificationLoss(Loss):
    pass


cdef class BinaryLoss(ClassificationLoss):
    """A loss appropriate for binary classification tasks."""

    cpdef double eval(self, bint y_true, double y_pred):
        """Returns the loss."""

    cpdef double gradient(self, bint y_true, double y_pred):
        """Returns the gradient with respect to ``y_pred``."""


cdef class MultiClassLoss(ClassificationLoss):
    """A loss appropriate for multi-class classification tasks."""

    cpdef double eval(self, object y_true, dict y_pred):
        """Returns the loss."""

    cpdef dict gradient(self, object y_true, dict y_pred):
        """Returns the gradient with respect to ``y_pred``."""


cdef class RegressionLoss(Loss):
    """A loss appropriate for regression tasks."""

    cpdef double eval(self, double y_true, double y_pred):
        """Returns the loss."""

    cpdef double gradient(self, double y_true, double y_pred):
        """Returns the gradient with respect to ``y_pred``."""


cdef class Absolute(RegressionLoss):
    """Absolute loss, also known as the mean absolute error or L1 loss.

    Mathematically, it is defined as

    .. math:: L = |p_i - y_i|

    It's gradient w.r.t. to $p_i$ is

    .. math:: \\frac{\\partial L}{\\partial p_i} = sgn(p_i - y_i)

    Example:

        ::

            >>> from creme import optim

            >>> loss = optim.losses.Absolute()
            >>> loss.eval(-42, 42)
            84.0
            >>> loss.gradient(1, 2)
            1.0
            >>> loss.gradient(2, 1)
            -1.0

    """

    cpdef double eval(self, double y_true, double y_pred):
        return abs(y_pred - y_true)

    cpdef double gradient(self, double y_true, double y_pred):
        if y_pred > y_true:
            return 1
        return -1


cdef class Cauchy(RegressionLoss):
    """Cauchy loss function.

    References:
        1. `Effect of MAE <https://www.kaggle.com/c/allstate-claims-severity/discussion/24520#140163>`_
        2. `Paris Madness <https://www.kaggle.com/raddar/paris-madness>`_

    """

    def __init__(self, C=80):
        self.C = C

    cpdef double eval(self, double y_true, double y_pred):
        return abs(y_pred - y_true)

    cpdef double gradient(self, double y_true, double y_pred):
        diff = y_pred - y_true
        return diff / ((diff / self.C) ** 2 + 1)


cdef class CrossEntropy(MultiClassLoss):
    """Cross entropy loss.

    This is a generalization of logistic loss to multiple classes.

    Example:

        ::

            >>> from creme import optim

            >>> y_true = [0, 1, 2, 2]
            >>> y_pred = [
            ...     {0: 0.29450637, 1: 0.34216758, 2: 0.36332605},
            ...     {0: 0.21290077, 1: 0.32728332, 2: 0.45981591},
            ...     {0: 0.42860913, 1: 0.33380113, 2: 0.23758974},
            ...     {0: 0.44941979, 1: 0.32962558, 2: 0.22095463}
            ... ]

            >>> loss = optim.losses.CrossEntropy()

            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     print(loss.eval(y_t, y_p))
            1.222454...
            1.116929...
            1.437209...
            1.509797...

            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     print(loss.gradient(y_t, y_p))
            {0: -0.70549363, 1: 0.34216758, 2: 0.36332605}
            {0: 0.21290077, 1: -0.67271668, 2: 0.45981591}
            {0: 0.42860913, 1: 0.33380113, 2: -0.76241026}
            {0: 0.44941979, 1: 0.32962558, 2: -0.77904537}

    References:
        1. `What is Softmax regression and how is it related to Logistic regression? <https://github.com/rasbt/python-machine-learning-book/blob/master/faq/softmax_regression.md>`_

    """

    cpdef double eval(self, object y_true, dict y_pred):
        cdef double total = 0

        for label, proba in y_pred.items():
            if y_true == label:
                total += math.log(clamp_proba(proba))

        return -total

    cpdef dict gradient(self, object y_true, dict y_pred):
        return {
            label: clamp_proba(y_pred.get(label, 0.)) - (y_true == label)
            for label in {*y_pred.keys(), y_true}
        }


cdef class Hinge(BinaryLoss):
    """Computes the hinge loss.

    Mathematically, it is defined as

    .. math:: L = max(0, 1 - p_i * y_i)

    It's gradient w.r.t. to $p_i$ is

    .. math::
        \\frac{\\partial L}{\\partial y_i} = \\left\{
        \\begin{array}{ll}
            \\ 0  &   p_iy_i \geqslant 1  \\\\
            \\ - y_i & p_iy_i < 1
        \\end{array}
        \\right.

    Example:

        ::

            >>> from creme import optim
            >>> import numpy as np
            >>> from sklearn import svm
            >>> from sklearn.metrics import hinge_loss

            >>> X = [[0], [1]]
            >>> y = [-1, 1]
            >>> lin_svm = svm.LinearSVC(random_state=0).fit(X, y)

            >>> y_true = [0, 1, 1]
            >>> y_pred = lin_svm.decision_function([[-2], [3], [0.5]])

            >>> hinge_loss([0, 1, 1], y_pred)
            0.303036...

            >>> loss = optim.losses.Hinge()
            >>> np.mean([loss.eval(y_t, y_p) for y_t, y_p in zip(y_true, y_pred)])
            0.303036...

    """

    cpdef double eval(self, bint y_true, double y_pred):
        # Our convention is to use 0s instead of -1s for negatives, but the Hinge loss uses -1s as
        # a convention
        y_true = y_true or -1
        return math.fmax(0, 1 - y_true * y_pred)

    cpdef double gradient(self, bint y_true, double y_pred):
        """Returns the gradient with respect to ``y_pred``.

        References:
            1. `Wolfram Alpha derivation <https://www.wolframalpha.com/input/?i=derivative+max(0,+1+-+p+*+y)+wrt+p>`_

        """
        y_true = y_true or -1
        if y_true * y_pred < 1:
            return -y_pred
        return 0


cdef class EpsilonInsensitiveHinge(RegressionLoss):

    cdef readonly double eps

    def __init__(self, eps=0.1):
        self.eps = eps

    cpdef double eval(self, double y_true, double y_pred):
        # Our convention is to use 0s instead of -1s for negatives, but the Hinge loss uses -1s as
        # a convention
        y_true = y_true or -1
        return math.fmax(0, math.fabs(y_pred - y_true) - self.eps)

    cpdef double gradient(self, double y_true, double y_pred):
        """Returns the gradient with respect to ``y_pred``.

        References:
            1. `Wolfram Alpha <https://www.wolframalpha.com/input/?i=derivative+max(0,+abs(p+-+y)+-+eps)+wrt+p>`_

        """
        y_true = y_true or -1
        if y_pred > y_true + self.eps:
            return 1
        elif y_pred + self.eps < y_true:
            return -1
        return 0


cdef class Log(BinaryLoss):
    """Logarithmic loss.

    This loss function expects each provided ``y_pred`` to be a log odds ratio, not a probability.

    """

    cpdef double eval(self, bint y_true, double y_pred):
        y_true = y_true or -1
        z = y_pred * y_true
        if z > 18.:
            return math.exp(-z)
        if z < -18.:
            return -z
        return math.log(1.0 + math.exp(-z))

    cpdef double gradient(self, bint y_true, double y_pred):
        y_true = y_true or -1
        z = y_pred * y_true
        if z > 18.:
            return math.exp(-z) * -y_true
        if z < -18.:
            return -y_true
        return -y_true / (math.exp(z) + 1.0)


cdef class Quantile(RegressionLoss):
    """Quantile loss.

    Parameters:
        alpha (float): Desired quantile to attain.

    Example:

        ::

            >>> from creme import optim

            >>> loss = optim.losses.Quantile(0.5)
            >>> loss.eval(1, 3)
            1.0

            >>> loss.gradient(1, 3)
            0.5

            >>> loss.gradient(3, 1)
            -0.5

    References:
        1. `Wikipedia article on quantile regression <https://www.wikiwand.com/en/Quantile_regression>`_
        2. `Derivative from WolframAlpha <https://www.wolframalpha.com/input/?i=derivative+(y+-+p)+*+(alpha+-+Boole(y+-+p))+wrt+p>`_

    """

    cdef readonly double alpha

    def __init__(self, alpha):
        self.alpha = alpha

    cpdef double eval(self, double y_true, double y_pred):
        diff = y_pred - y_true
        return (self.alpha - (diff < 0)) * diff

    cpdef double gradient(self, double y_true, double y_pred):
        return (y_true < y_pred) - self.alpha


cdef class Squared(RegressionLoss):
    """Computes the squared loss, also known as the L2 loss.

    Mathematically, it is defined as

    .. math:: L = (p_i - y_i) ^ 2

    It's gradient w.r.t. to $p_i$ is

    .. math:: \\frac{\\partial L}{\\partial p_i} = 2 \times (p_i - y_i)

    One thing to note is that this convention is consistent with Vowpal Wabbit and PyTorch, but
    not with scikit-learn. Indeed scikit-learn divides the loss by 2, making the 2 dissapear in
    the gradient.

    Example:

        ::

            >>> from creme import optim

            >>> loss = optim.losses.Squared()
            >>> loss.eval(-4, 5)
            81.0
            >>> loss.gradient(-4, 5)
            18.0
            >>> loss.gradient(5, -4)
            -18.0

    """

    cpdef double eval(self, double y_true, double y_pred):
        return (y_pred - y_true) * (y_pred - y_true)

    cpdef double gradient(self, double y_true, double y_pred):
        return 2. * (y_pred - y_true)



import inspect
import types


__all__ = ['pretty_format_class']


def pretty_format_class(cls):
    """Returns a pretty representation of a class."""

    rep = f'{cls.__class__.__name__} ('
    init = inspect.signature(cls.__init__)

    for name, param in init.parameters.items():

        # Retrieve the attribute associated with the parameter
        if param.default is None or param.default == param.empty:
            try:
                attr = getattr(cls, name)
            except AttributeError:
                continue
        else:
            attr = param.default

        # Prettify the attribute when applicable
        if isinstance(attr, str):
            attr = f"'{attr}'"
        elif isinstance(attr, types.FunctionType):
            attr = attr.__name__
        elif isinstance(attr, set):
            attr = sorted(attr)

        rep += f'\n    {name}={attr}'

    if init.parameters:
        rep += '\n)'
    else:
        rep += ')'

    return rep



import collections

from . import base


__all__ = ['Multinomial']


class Multinomial(collections.Counter, base.DiscreteDistribution):
    """Multinomial distribution for categorical data.

    Example:

        ::

            >>> from creme import proba

            >>> p = proba.Multinomial(['green'] * 3)
            >>> p = p.update('red')

            >>> p.pmf('red')
            0.25

            >>> p.update('red').update('red').pmf('green')
            0.5

    """

    def __init__(self, events=None):
        super().update(events)
        self._n = sum(self.values())

    @property
    def n_samples(self):
        return self._n

    def update(self, x):
        super().update([x])
        self._n += 1
        return self

    def pmf(self, x):
        try:
            return self[x] / self._n
        except ZeroDivisionError:
            return 0.

    def __str__(self):
        return '\n'.join(f'P({c}) = {self.pmf(c):.3f}' for c in self)



import abc
import numbers
try:
    import graphviz
    GRAPHVIZ_INSTALLED = True
except ImportError:
    GRAPHVIZ_INSTALLED = False

from .. import base
from .. import proba

from . import criteria
from . import leaf
from . import splitting


CRITERIA_CLF = {'gini': criteria.gini_impurity, 'entropy': criteria.entropy}


class BaseDecisionTree(abc.ABC):

    def __init__(self, criterion='gini', patience=250, max_depth=5, min_split_gain=0.,
                 min_child_samples=20, confidence=1e-10, tie_threshold=5e-2, n_split_points=30,
                 max_bins=60):
        self.criterion = CRITERIA_CLF[criterion]
        self.patience = patience
        self.max_depth = max_depth
        self.min_split_gain = min_split_gain
        self.min_child_samples = min_child_samples
        self.confidence = confidence
        self.tie_threshold = tie_threshold
        self.n_split_points = n_split_points
        self.max_bins = max_bins

        self.root = leaf.Leaf(depth=0, tree=self, target_dist=proba.Multinomial())

    def fit_one(self, x, y):
        self.root = self.root.update(x, y)
        return self

    @abc.abstractmethod
    def _get_split_enum(self, typ):
        """Returns the appropriate split enumerator for a given type."""

    def draw(self):
        """Returns a GraphViz representation of the decision tree."""

        if not GRAPHVIZ_INSTALLED:
            raise RuntimeError('graphviz is not installed')

        dot = graphviz.Digraph()

        def add_node(node, path):

            if isinstance(node, leaf.Leaf):
                # Draw a leaf
                dot.node(path, str(node.target_dist), shape='box')
            else:
                # Draw a branch
                dot.node(path, str(node.split))
                add_node(node.left, f'{path}0')
                add_node(node.right, f'{path}1')

            # Draw the link with the previous node
            is_root = len(path) == 1
            if not is_root:
                dot.edge(path[:-1], path)

        add_node(node=self.root, path='0')

        return dot

    def debug_one(self, x):
        """Prints an explanation of how ``x`` is predicted."""
        node = self.root

        while isinstance(node, leaf.Branch):
            if node.split.test(x):
                print('not', node.split)
                node = node.left
            else:
                print(node.split)
                node = node.right

        print(node.target_dist)


class DecisionTreeClassifier(BaseDecisionTree, base.MultiClassifier):
    """Decision tree classifier.

    Parameters:
        criterion (str): The function to measure the quality of a split. Set to ``'gini'`` in order
            to use Gini impurity and ``'entropy'`` for information gain.
        patience (int): Time to wait between split attempts.
        max_depth (int): Maximum tree depth.
        min_split_gain (float): Minimum impurity gain required to make a split eligible.
        min_child_samples (int): Minimum number of data needed in a leaf.
        confidence (float): Threshold used to compare with the Hoeffding bound.
        tie_threshold (float): Threshold to handle ties between equally performing attributes.
        n_split_points (int): Number of split points considered for splitting numerical variables.
        max_bins (int): Number of histogram bins used for approximating the distribution of
            numerical variables.

    Attributes:
        root (Leaf)

    Example:

        ::

            >>> from creme import datasets
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import tree

            >>> X_y = datasets.fetch_electricity()

            >>> model = tree.DecisionTreeClassifier(
            ...     patience=2000,
            ...     confidence=1e-5,
            ...     criterion='gini'
            ... )

            >>> metric = metrics.LogLoss()

            >>> model_selection.online_score(X_y, model, metric)
            LogLoss: 0.55375

    References:
        1. `Mining High-Speed Data Streams <https://homes.cs.washington.edu/~pedrod/papers/kdd00.pdf>`_
        2. `The Morning Paper <https://blog.acolyer.org/2015/08/26/mining-high-speed-data-streams/>`_

    """

    def _get_split_enum(self, name, value):
        if isinstance(value, numbers.Number):
            return splitting.NumericSplitEnum(
                feature_name=name,
                n_bins=self.max_bins,
                n_splits=self.n_split_points
            )

        elif isinstance(value, str):
            return splitting.CategoricalSplitEnum(feature_name=name)

        raise ValueError(f'Unsupported feature type: {type(value)} ({name}: {value})')

    def predict_proba_one(self, x):
        return self.root.get_leaf(x).predict(x)



import importlib
import inspect

import pytest


@pytest.mark.parametrize(
    'func',
    [
        pytest.param(func, id=name)
        for name, func
        in inspect.getmembers(importlib.import_module(f'creme.datasets'), inspect.isfunction)
        if name.startswith('fetch')
    ]
)
@pytest.mark.web
def test_fetch(func):
    X_y = func()
    x, y = next(X_y)


@pytest.mark.parametrize(
    'func',
    [
        pytest.param(func, id=name)
        for name, func
        in inspect.getmembers(importlib.import_module(f'creme.datasets'), inspect.isfunction)
        if name.startswith('load')
    ]
)
def test_load(func):
    X_y = func()
    x, y = next(X_y)



"""Anomaly detection.

The estimators in ``anomaly`` are slightly different than the rest of the estimators. Instead of a
``predict_one`` method, each anomaly detector has a ``score_one`` method which returns an anomaly
score for a given set of features. Anomalies will have high scores whereas normal observations will
have low scores. The range of the scores depends on the estimator.

"""
from .hst import HalfSpaceTrees


__all__ = ['HalfSpaceTrees']



"""Online feature selection."""
from .k_best import SelectKBest
from .random import RandomDiscarder
from .variance import VarianceThreshold


__all__ = [
    'RandomDiscarder',
    'SelectKBest',
    'VarianceThreshold'
]



import numpy as np

from .. import base


class RandomDiscarder(base.Transformer):
    """Transformer that randomly discards features.

    Example:

        ::

            >>> from creme import feature_selection
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X, _ = datasets.make_regression(n_samples=30, n_features=8)
            >>> discarder = feature_selection.RandomDiscarder(n_to_keep=5)

            >>> for x, _ in stream.iter_array(X):
            ...     x_pruned = discarder.transform_one(x)
            ...     assert len(x_pruned) == 5
            ...     # creme's transformers are pure, so the input should not change
            ...     assert len(x) == 8

    """

    def __init__(self, n_to_keep):
        self.n_to_keep = n_to_keep

    def transform_one(self, x):

        if len(x) - self.n_to_keep > 0:
            return {
                i: x[i]
                for i in np.random.choice(list(x.keys()), size=self.n_to_keep, replace=False)
            }

        return x



from .. import optim
from .. import stats

from . import base


__all__ = ['CrossEntropy']


class BaseCrossEntropy(base.MultiClassMetric):

    @property
    def bigger_is_better(self):
        return False

    @property
    def requires_labels(self):
        return False


class CrossEntropy(stats.Mean, BaseCrossEntropy):
    """Multiclass generalization of the logarithmic loss.

    Example:

        ::

            >>> from creme import metrics
            >>> from sklearn.metrics import log_loss

            >>> y_true = [0, 1, 2, 2]
            >>> y_pred = [
            ...     {0: 0.29450637, 1: 0.34216758, 2: 0.36332605},
            ...     {0: 0.21290077, 1: 0.32728332, 2: 0.45981591},
            ...     {0: 0.42860913, 1: 0.33380113, 2: 0.23758974},
            ...     {0: 0.44941979, 1: 0.32962558, 2: 0.22095463}
            ... ]

            >>> metric = metrics.CrossEntropy()

            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     metric = metric.update(y_t, y_p)
            ...     print(metric.get())
            1.222454...
            1.169691...
            1.258864...
            1.321597...

            >>> metric
            CrossEntropy: 1.321598

    """

    def update(self, y_true, y_pred, sample_weight=1.):
        return super().update(x=optim.losses.CrossEntropy().eval(y_true, y_pred), w=sample_weight)

    def revert(self, y_true, y_pred, sample_weight=1.):
        return super().revert(x=optim.losses.CrossEntropy().eval(y_true, y_pred), w=sample_weight)



import collections
import functools

import numpy as np
from sklearn import utils

from .. import stats
from .. import optim

from . import base


__all__ = ['SVD']


class SVD(base.Recommender):
    """ SVD implementation for recommender systems

    Here is an implementation of the Funk SVD, the code is based on the SVD from ``surprise``
    but you can choose the optimizer (in ``surprise``, only SGD is available)

    Parameters:
        n_factors (default=0): The number of latent factors to compute for each vector
        mu (default=0): The mean of the normal distribution for the random weights initialisation
        sigma (default=0.1): The standard deviation of the normal distribution for the random
            weights initialisation
        row_optimizer (optim.Optimizer): Optimizer used to tune the rows weights.
        col_optimizer (optim.Optimizer): Optimizer used to tune the column weights.
        loss (optim.Loss): Loss function to optimize for.
        l2 (float, default=0): regularization amount used to push weights towards 0.
        global_mean (stats.Mean)
        random_state (int, ``numpy.random.RandomState`` instance or None): If int, ``random_state``
            is the seed used by the random number generator; if ``RandomState`` instance,
            ``random_state`` is the random number generator; if ``None``, the random number
            generator is the ``RandomState`` instance used by `numpy.random`.

    Attributes:
        row_biases (collections.defaultdict): The biases of the rows elements (users)
        col_biases (collections.defaultdict): The biases of the columns elements (items)
        row_vec (collections.defaultdict): The latents vectors of the rows elements
            randomly initialized (users)
        col_vec (collections.defaultdict): The latents vectors of the columns elements
            randomly initialized (items)

    Example:

        ::

            >>> from creme import reco

            >>> ratings = [
            ...     ('Alice', 'Superman', 8),
            ...     ('Alice', 'Terminator', 9),
            ...     ('Alice', 'Star Wars', 8),
            ...     ('Alice', 'Notting Hill', 2),
            ...     ('Alice', 'Harry Potter ', 5),
            ...     ('Bob', 'Superman', 8),
            ...     ('Bob', 'Terminator', 9),
            ...     ('Bob', 'Star Wars', 8),
            ...     ('Bob', 'Notting Hill', 2)
            ... ]

            >>> model = reco.SVD(
            ...     n_factors=10,
            ...     row_optimizer=optim.SGD(0.005),
            ...     col_optimizer=optim.SGD(0.005),
            ...     random_state=42
            ... )

            >>> for user, movie, rating in ratings:
            ...     _ = model.fit_one(user, movie, rating)

            >>> model.predict_one('Bob', 'Harry Potter')
            6.552539...

    """

    def __init__(self, n_factors=20, mu=0, sigma=0.1, row_optimizer=None, col_optimizer=None,
                 loss=None, l2=0., random_state=None):

        self.n_factors = n_factors
        self.row_optimizer = optim.SGD() if row_optimizer is None else row_optimizer
        self.col_optimizer = optim.SGD() if col_optimizer is None else row_optimizer
        self.loss = optim.losses.Squared() if loss is None else row_optimizer
        self.l2 = l2
        self.global_mean = stats.Mean()
        self.random_state = utils.check_random_state(random_state)
        self.row_biases = collections.defaultdict(float)
        self.col_biases = collections.defaultdict(float)
        random_weights = functools.partial(
            self.random_state.normal,
            loc=mu,
            scale=sigma,
            size=n_factors
        )
        self.row_vec = collections.defaultdict(random_weights)
        self.col_vec = collections.defaultdict(random_weights)

    def fit_one(self, r_id, c_id, y):

        self.row_optimizer.update_before_pred(w=self.row_biases)
        self.col_optimizer.update_before_pred(w=self.col_biases)

        # Predict the value
        y_pred = self.predict_one(r_id, c_id)

        # Compute the gradient of the loss with respect to the prediction
        loss_gradient = self.loss.gradient(y, y_pred)

        # Update row bias
        r_grad_bias = {r_id: loss_gradient + self.l2 * self.col_biases[r_id]}
        self.row_biases = self.row_optimizer.update_after_pred(self.row_biases, r_grad_bias)

        # Update the row vector
        r_grad_vec = {r_id: loss_gradient * self.col_vec[c_id] + self.l2 * self.row_vec[r_id]}
        self.row_vec = self.row_optimizer.update_after_pred(self.row_vec, r_grad_vec)

        # Update column bias
        c_grad_bias = {c_id: loss_gradient + self.l2 * self.col_biases[c_id]}
        self.col_biases = self.col_optimizer.update_after_pred(self.col_biases, c_grad_bias)

        # Update the column vector
        c_grad_vec = {c_id: loss_gradient * self.row_vec[r_id] + self.l2 * self.col_vec[c_id]}
        self.col_vec = self.col_optimizer.update_after_pred(self.col_vec, c_grad_vec)

        # Update the global mean
        self.global_mean.update(y)

        return y_pred

    def predict_one(self, r_id, c_id):

        # Initialize the prediction to the mean
        y_pred = (self.global_mean.get() or 0)

        # Add the user bias if the user is known
        y_pred += self.row_biases.get(r_id, 0)

        # Add the item bias if the item is known
        y_pred += self.col_biases.get(c_id, 0)

        # Add the dot product of the row and the column if both are known
        y_pred += np.dot(self.row_vec.get(r_id, np.zeros(self.n_factors)),
                         self.col_vec.get(c_id, np.zeros(self.n_factors)))

        return y_pred



import math

from sklearn import utils

from . import base


class NUnique(base.Univariate):
    """Approximate number of unique values counter.

    This is basically an implementation of the HyperLogLog algorithm. Adapted from
    `here <https://github.com/clarkduvall/hypy>`_. The code is a bit too terse but it will do for
    now.

    Parameters:
        error_rate (float): Desired error rate. Memory usage is inversely proportional to this
            value.
        encoding (str): Encoding used for `sklearn.utils.murmurhash3_32`. Defaults to ``'utf-8'``.
        random_state (int): Integer seed for the hashing algorithm. Defaults to `None`.

    Attributes:
        n_bits (int)
        n_buckets (int)
        buckets (list)

    Example:

        ::

            >>> import string
            >>> from creme import stats

            >>> alphabet = string.ascii_lowercase
            >>> n_unique = stats.NUnique(error_rate=0.1)

            >>> n_unique.update('a').get()
            1

            >>> n_unique.update('b').get()
            2

            >>> for letter in alphabet:
            ...     n_unique = n_unique.update(letter)
            >>> n_unique.get()
            26

    References:

        1. `My favorite algorithm (and data structure): HyperLogLog <https://odino.org/my-favorite-data-structure-hyperloglog/>`_
        2. `HyperLogLog: the analysis of a near-optimal cardinality estimation algorithm <http://algo.inria.fr/flajolet/Publications/FlFuGaMe07.pdf>`_

    """

    P32 = 2 ** 32

    def __init__(self, error_rate=0.01, encoding='utf-8', random_state=None):
        self.n_bits = int(math.ceil(math.log((1.04 / error_rate) ** 2, 2)))
        self.n_buckets = 1 << self.n_bits
        self.buckets = [0] * self.n_buckets
        self.encoding = encoding
        self.random_state = utils.check_random_state(random_state)
        self.seed = self.random_state.randint(0, 2 ** 32 - 1)

    @property
    def name(self):
        return 'n_unique'

    def _hash_str(self, string):
        return utils.murmurhash3_32(bytes(string, self.encoding), self.seed)

    def update(self, x):
        x = self._hash_str(x)
        i = x & NUnique.P32 - 1 >> 32 - self.n_bits
        z = 35 - len(bin(NUnique.P32 - 1 & x << self.n_bits | 1 << self.n_bits - 1))
        self.buckets[i] = max(self.buckets[i], z)
        return self

    def get(self):
        a = ({16: 0.673, 32: 0.697, 64: 0.709}[self.m]
             if self.n_buckets <= 64 else 0.7213 / (1 + 1.079 / self.n_buckets))
        e = a * self.n_buckets * self.n_buckets / sum(1.0 / (1 << x) for x in self.buckets)
        if e <= self.n_buckets * 2.5:
            z = len([r for r in self.buckets if not r])
            return int(self.n_buckets * math.log(float(self.n_buckets) / z) if z else e)
        return int(e if e < NUnique.P32 / 30 else -NUnique.P32 * math.log(1 - e / NUnique.P32))



from .. import base


__all__ = ['Whitelister']


class Whitelister(base.Transformer):
    """Subsets a set of features by applying a whitelist.

    Parameters:
        whitelist (strs): Key(s) to keep.

    Example:

        ::

            >>> from creme import compose

            >>> x = {'a': 42, 'b': 12}
            >>> compose.Whitelister('a', 'zoidberg').transform_one(x)
            {'a': 42}

            >>> compose.Whitelister('b').transform_one(x)
            {'b': 12}

    """

    def __init__(self, *whitelist):
        self.whitelist = set(whitelist)

    def transform_one(self, x):
        return {i: x[i] for i in set(x.keys()) & self.whitelist}

    def __str__(self):
        return str(sorted(self.whitelist))



"""Compatibility with other libraries.

This module contains wrappers for making ``creme`` estimators compatible with other libraries, and
vice-versa whenever possible.

"""
from .sklearn import convert_creme_to_sklearn
from .sklearn import convert_sklearn_to_creme
from .sklearn import CremeClassifierWrapper
from .sklearn import CremeRegressorWrapper
from .sklearn import SKLRegressorWrapper
from .sklearn import SKLClassifierWrapper
from .sklearn import SKLClustererWrapper
from .sklearn import SKLTransformerWrapper


__all__ = [
    'convert_creme_to_sklearn',
    'convert_sklearn_to_creme',
    'CremeClassifierWrapper',
    'CremeRegressorWrapper',
    'SKLRegressorWrapper',
    'SKLClassifierWrapper',
    'SKLClustererWrapper',
    'SKLTransformerWrapper'
]



from .. import utils

from . import base
from . import per_class


__all__ = ['Rolling']


class Rolling(base.WrapperMetric, utils.Window):
    """Wrapper for computing metrics over a window.

    Parameters:
        metric (metrics.Metric)

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = [3, -0.5, 2, 7]
            >>> y_pred = [2.5, 0.0, 2, 8]

            >>> metric = metrics.Rolling(metrics.MSE(), window_size=2)

            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     print(metric.update(y_t, y_p))
            Rolling of size 2 MSE: 0.25
            Rolling of size 2 MSE: 0.25
            Rolling of size 2 MSE: 0.125...
            Rolling of size 2 MSE: 0.5...

    """

    def __init__(self, metric, window_size):
        super().__init__(size=window_size)
        self.window_size = window_size
        self._metric = metric

    @property
    def metric(self):
        return self._metric

    def update(self, y_true, y_pred, sample_weight=1.):
        self.metric.update(y_true, y_pred, sample_weight)
        if len(self) == self.window_size:
            self.metric.revert(*self[0])
        self.append((y_true, y_pred, sample_weight))
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self.metric.revert(y_true, y_pred, sample_weight)
        return self

    def get(self):
        return self.metric.get()

    def __str__(self):
        if isinstance(self._metric, per_class.PerClass):
            return f'Rolling of size {self.window_size}\n' + '\n'.join((
                f'    {s}'
                for s in str(self.metric).split('\n'))
            )
        return f'Rolling of size {self.window_size} {str(self.metric)}'



import collections
import copy
import functools
import typing

from .. import base


__all__ = ['Agg', 'TargetAgg']


class Agg(base.Transformer):
    """Computes a streaming aggregate.

    At each step, the running statistic ``how`` of group ``by`` is updated with the value of
    ``on``. You can combine this with a ``creme.compose.TransformerUnion`` to extract many
    aggregate statistics in one go. The keys of the ``dict`` containing the aggregates is
    automatically guessed from ``how``, ``by``, and ``on``.

    Parameters:
        on (str): The feature on which to compute.
        by (str): The feature on which to group.
        how (stats.Univariate): The statistic to compute.

    Attributes:
        groups (collections.defaultdict): Maps group keys to univariate statistics.
        feature_name (str): The name of the feature in the output.

    Example:

        ::

            >>> from creme import feature_extraction
            >>> from creme import stats

            >>> X = [
            ...     {'place': 'Taco Bell', 'revenue': 42},
            ...     {'place': 'Burger King', 'revenue': 16},
            ...     {'place': 'Burger King', 'revenue': 24},
            ...     {'place': 'Taco Bell', 'revenue': 58},
            ...     {'place': 'Burger King', 'revenue': 20},
            ...     {'place': 'Taco Bell', 'revenue': 50}
            ... ]

            >>> agg = feature_extraction.Agg(
            ...     on='revenue',
            ...     by='place',
            ...     how=stats.Mean()
            ... )

            >>> for x in X:
            ...     print(agg.fit_one(x).transform_one(x))
            {'revenue_mean_by_place': 42.0}
            {'revenue_mean_by_place': 16.0}
            {'revenue_mean_by_place': 20.0}
            {'revenue_mean_by_place': 50.0}
            {'revenue_mean_by_place': 20.0}
            {'revenue_mean_by_place': 50.0}

    References:
        1. `Streaming groupbys in pandas for big datasets <(https://maxhalford.github.io/blog/streaming-groupbys-in-pandas-for-big-datasets/>`_

    """

    def __init__(self, on, by, how):
        self.on = on
        self.by = by if isinstance(by, (typing.List, typing.Tuple)) else [by]
        self.how = how
        self.groups = collections.defaultdict(functools.partial(copy.deepcopy, how))
        self.feature_name = f'{self.on}_{self.how.name}_by_{"_and_".join(self.by)}'

    def _get_key(self, x):
        return '_'.join(str(x[k]) for k in self.by)

    def fit_one(self, x, y=None):
        self.groups[self._get_key(x)].update(x[self.on])
        return self

    def transform_one(self, x):
        return {self.feature_name: self.groups[self._get_key(x)].get()}

    def __str__(self):
        return self.feature_name


class TargetAgg(base.Transformer):
    """Computes a streaming group by on the target.

    At each step, the running statistic ``how`` of group ``by`` is updated with the target.

    Parameters:
        by (str): The feature on which to group.
        how (stats.Univariate): The statistic to compute.
        target_name (str): Name used in the result.

    Attributes:
        groups (dict): Maps group keys to univariate statistics.
        feature_name (str): The name of the feature in the output.

    Example:

        ::

            >>> import creme

            >>> X = [
            ...     {'place': 'Taco Bell', 'revenue': 42},
            ...     {'place': 'Burger King', 'revenue': 16},
            ...     {'place': 'Burger King', 'revenue': 24},
            ...     {'place': 'Taco Bell', 'revenue': 58},
            ...     {'place': 'Burger King', 'revenue': 20},
            ...     {'place': 'Taco Bell', 'revenue': 50}
            ... ]

            >>> agg = creme.feature_extraction.TargetAgg(
            ...     by='place',
            ...     how=creme.stats.BayesianMean(
            ...         prior=3,
            ...         prior_weight=1
            ...     )
            ... )

            >>> for x in X:
            ...     print(agg.transform_one(x))
            ...     y = x.pop('revenue')
            ...     agg = agg.fit_one(x, y)
            {'target_bayes_mean_by_place': 3.0}
            {'target_bayes_mean_by_place': 3.0}
            {'target_bayes_mean_by_place': 9.5}
            {'target_bayes_mean_by_place': 22.5}
            {'target_bayes_mean_by_place': 14.333333...}
            {'target_bayes_mean_by_place': 34.333333...}

    References:
        1. `Streaming groupbys in pandas for big datasets <(https://maxhalford.github.io/blog/streaming-groupbys-in-pandas-for-big-datasets/>`_

    """

    def __init__(self, by, how, target_name='target'):
        self.by = by if isinstance(by, (typing.List, typing.Tuple)) else [by]
        self.how = how
        self.target_name = target_name
        self.groups = collections.defaultdict(functools.partial(copy.deepcopy, how))
        self.feature_name = f'{target_name}_{how.name}_by_{"_and_".join(self.by)}'

    def is_supervised(self):
        return True

    def _get_key(self, x):
        return '_'.join(str(x[k]) for k in self.by)

    def fit_one(self, x, y=None):
        self.groups[self._get_key(x)].update(y)
        return self

    def transform_one(self, x):
        return {self.feature_name: self.groups[self._get_key(x)].get()}

    def __str__(self):
        return self.feature_name



import math

from ..proba.base import ContinuousDistribution


class Branch:

    def __init__(self, split, left, right, tree):
        self.split = split
        self.left = left
        self.right = right
        self.tree = tree

    @property
    def size(self):
        return self.left.size + self.right.size

    def get_leaf(self, x):
        if self.split(x):
            return self.left.get_leaf(x)
        return self.right.get_leaf(x)

    def update(self, x, y):
        if self.split(x):
            self.left = self.left.update(x, y)
            return self
        self.right = self.right.update(x, y)
        return self


class Leaf:

    def __init__(self, depth, tree, target_dist):
        self.depth = depth
        self.tree = tree
        self.target_dist = target_dist
        self.n_samples = 0
        self.split_enums = {}

    @property
    def size(self):
        return 1

    @property
    def n_classes(self):
        """The number of observed classes."""
        if isinstance(self.target_dist, ContinuousDistribution):
            raise ValueError('The target is continuous, hence there are not classes')
        return len(self.target_dist)

    @property
    def is_pure(self):
        try:
            return self.n_classes < 2
        except ValueError:
            return False

    def get_leaf(self, x):
        return self

    @property
    def hoeffding_bound(self):
        """Returns the current Hoeffding bound.

        TODO: handle continuous target
        """
        R = math.log(self.n_classes)
        n = self.n_samples
        δ = self.tree.confidence
        return math.sqrt(R ** 2 * math.log(1 / δ) / (2 * n))

    def update(self, x, y):

        # Update the class counts
        self.target_dist.update(y)
        self.n_samples += 1

        # Update the sufficient statistics of each feature's split searcher
        for i, xi in x.items():
            try:
                ss = self.split_enums[i]
            except KeyError:
                ss = self.split_enums[i] = self.tree._get_split_enum(name=i, value=xi)
            ss.update(x=xi, y=y)

        # Check if splitting is authorized or not
        if (
            self.depth >= self.tree.max_depth or
            self.is_pure or
            self.n_samples % self.tree.patience != 0
        ):
            return self

        # Search for the best split given the current information
        top_2_diff, split, left_dist, right_dist = self.find_best_split()

        # Calculate the Hoeffding bound
        ε = self.hoeffding_bound

        if top_2_diff > ε or ε < self.tree.tie_threshold:
            return Branch(
                split=split,
                left=Leaf(depth=self.depth + 1, tree=self.tree, target_dist=left_dist),
                right=Leaf(depth=self.depth + 1, tree=self.tree, target_dist=right_dist),
                tree=self.tree
            )
        return self

    def find_best_split(self):
        """Returns the best potential split."""

        current_impurity = self.tree.criterion(dist=self.target_dist)
        best_gain = -math.inf
        second_best_gain = -math.inf
        best_split = None
        best_l_dist = None
        best_r_dist = None

        # For each feature
        for ss in self.split_enums.values():

            # For each candidate split
            for split, l_dist, r_dist in ss.enumerate_splits(target_dist=self.target_dist):

                # Ignore the split if it results in a new leaf with not enough samples
                if (
                    l_dist.n_samples < self.tree.min_child_samples or
                    r_dist.n_samples < self.tree.min_child_samples
                ):
                    continue

                # Compute the gain brought by the split
                left_impurity = self.tree.criterion(dist=l_dist)
                right_impurity = self.tree.criterion(dist=r_dist)
                impurity = l_dist.n_samples * left_impurity + r_dist.n_samples * right_impurity
                impurity /= l_dist.n_samples + r_dist.n_samples
                gain = current_impurity - impurity

                # Ignore the split if the gain in impurity is too low
                if gain < self.tree.min_split_gain:
                    continue

                # Check if the gain brought by the candidate split is better than the current best
                if gain > best_gain:
                    best_gain, second_best_gain = gain, best_gain
                    best_split = split
                    best_l_dist = l_dist
                    best_r_dist = r_dist
                elif gain > second_best_gain:
                    second_best_gain = gain

        if best_split is None:
            raise RuntimeError('No best split was found')

        return best_gain - second_best_gain, best_split, best_l_dist, best_r_dist

    def predict(self, x):
        if isinstance(self.target_dist, ContinuousDistribution):
            return self.target_dist.mode
        return {c: self.target_dist.pmf(c) for c in self.target_dist}

    def predict_naive_bayes(self, x):

        y_pred = self.predict(x)

        for i, xi in x.items():
            if i in self.split_enums:
                for label, dist in self.split_enums[i].items():
                    if isinstance(dist, ContinuousDistribution):
                        y_pred[label] *= dist.pdf(xi)
                    else:
                        y_pred[label] *= dist.pmf(xi)

        total = sum(y_pred.values())

        if total == 0:
            return {label: 1. / len(y_pred) for label in y_pred}

        for label, proba in y_pred.items():
            y_pred[label] /= total

        return y_pred



"""Models composition."""
from .blacklist import Blacklister
from .func import FuncTransformer
from .pipeline import Pipeline
from .rename import Renamer
from .target_modifier import BoxCoxTransformRegressor
from .target_modifier import TargetModifierRegressor
from .union import TransformerUnion
from .whitelist import Whitelister


__all__ = [
    'Blacklister',
    'BoxCoxTransformRegressor',
    'FuncTransformer',
    'Pipeline',
    'Renamer',
    'TargetModifierRegressor',
    'TransformerUnion',
    'Whitelister'
]



from . import base


class Covariance(base.Bivariate):
    """Running covariance.

    Parameters:
        ddof (int): Delta Degrees of Freedom.

    Example:

        >>> from creme import stats

        >>> x = [-2.1,  -1,  4.3]
        >>> y = [   3, 1.1, 0.12]

        >>> cov = stats.Covariance()

        >>> for xi, yi in zip(x, y):
        ...     print(cov.update(xi, yi).get())
        0.0
        -1.044999...
        -4.286

    References:

        1. `Algorithms for calculating variance <https://www.wikiwand.com/en/Algorithms_for_calculating_variance#/Covariance>`_

    """

    def __init__(self, ddof=1):
        self.ddof = ddof
        self.mean_x = 0
        self.mean_y = 0
        self.n = 0
        self.C = 0

    def update(self, x, y):
        self.n += 1
        dx = x - self.mean_x
        self.mean_x += dx / self.n
        self.mean_y += (y - self.mean_y) / self.n
        self.C += dx * (y - self.mean_y)
        return self

    def get(self):
        return self.C / max(1, self.n - self.ddof)



from .. import ensemble

from . import tree


__all__ = ['RandomForestClassifier']


class RandomForestClassifier(ensemble.BaggingClassifier):
    """Random forest classifier.

    This is a thin wrapper over `ensemble.BaggingClassifier` and `tree.DecisionTreeClassifier`.

    Parameters:
        n_trees (int): The number of trees in the forest.
        random_state (int, ``numpy.random.RandomState`` instance or None): If int, ``random_state``
            is the seed used by the random number generator; if ``RandomState`` instance,
            ``random_state`` is the random number generator; if ``None``, the random number
            generator is the ``RandomState`` instance used by `numpy.random`.
        **tree_params (dict): The parameters of the decision tree.

    Example:

        ::

            >>> from creme import datasets
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import tree

            >>> X_y = datasets.fetch_electricity()

            >>> model = tree.RandomForestClassifier(
            ...     n_trees=5,
            ...     random_state=42,
            ...     # Tree parameters
            ...     patience=2000,
            ...     confidence=1e-5,
            ...     criterion='gini'
            ... )

            >>> metric = metrics.LogLoss()

            >>> model_selection.online_score(X_y, model, metric)
            LogLoss: 0.529859

    """

    def __init__(self, n_trees=10, random_state=None, **tree_params):
        super().__init__(
            model=tree.DecisionTreeClassifier(**tree_params),
            n_models=n_trees,
            random_state=random_state
        )

    def __str__(self):
        return 'RandomForestClassifier'



"""Helper functions for streaming data."""
import csv
import datetime as dt
import functools
import gzip
import itertools
import random
import types
import os

import numpy as np
from sklearn import utils

__all__ = [
    'iter_csv',
    'iter_array',
    'iter_pandas',
    'iter_sklearn_dataset',
    'iter_vaex',
    'simulate_qa',
    'shuffle'
]


def iter_array(X, y=None, feature_names=None, target_names=None, shuffle=False, random_state=None):
    """Yields rows from an array of features and an array of targets.

    This method is compatible with ``numpy`` arrays as well as Python lists.

    Parameters:
        X (array-like of shape (n_samples, n_features))
        y (array-like of shape (n_samples,))
        feature_names (list of length n_features)
        target_names (list of length n_outputs)
        shuffle (bool): Whether to shuffle the inputs or not.
        random_state (int, np.random.RandomState instance or None, default=None): If int,
            ``random_state`` is the seed used by the random number generator; if ``RandomState``
            instance, ``random_state`` is the random number generator; if ``None``, the random
            number generator is the ``RandomState`` instance used by ``np.random``.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    """
    feature_names = list(range(len(X[0]))) if feature_names is None else feature_names

    multioutput = y is not None and not np.isscalar(y[0])
    if multioutput and target_names is None:
        target_names = list(range(len(y[0])))

    # Shuffle the data
    rng = utils.check_random_state(random_state)
    if shuffle:
        order = rng.permutation(len(X))
        X, y = X[order], y if y is None else y[order]

    if multioutput:

        for x, yi in itertools.zip_longest(X, y if hasattr(y, '__iter__') else []):
            yield dict(zip(feature_names, x)), dict(zip(target_names, yi))

    else:

        for x, yi in itertools.zip_longest(X, y if hasattr(y, '__iter__') else []):
            yield dict(zip(feature_names, x)), yi


def iter_sklearn_dataset(dataset, **kwargs):
    """Yields rows from one of the datasets provided by scikit-learn.

    Parameters:
        dataset (sklearn.utils.Bunch): A scikit-learn dataset.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    """
    kwargs['X'] = dataset.data
    kwargs['y'] = dataset.target
    try:
        kwargs['feature_names'] = dataset.feature_names
    except AttributeError:
        pass

    for x, yi in iter_array(**kwargs):
        yield x, yi


def iter_pandas(X, y=None, **kwargs):
    """Yields rows from a ``pandas.DataFrame``.

    Parameters:
        X (pandas.DataFrame)
        y (array-like of shape (n_samples,))

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    """
    import pandas as pd

    kwargs['feature_names'] = X.columns
    if isinstance(y, pd.DataFrame):
        kwargs['target_names'] = y.columns

    for x, yi in iter_array(X.to_numpy(), y, **kwargs):
        yield x, yi


def iter_vaex(X, y=None, features=None, **kwargs):
    """Yields rows from a ``vaex.DataFrame``.

    Parameters:
        X (vaex.DataFrame): A vaex DataFrame housing the training featuers.
        y (string or vaex.Expression): The column or expression containing the target variable.
        features (list of strings or vaex.Expressions): A list of features used for training.
        If None, all columns in ``X`` will be used. Features specifying in ``y`` are ignored.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.
    """

    from vaex.utils import _ensure_strings_from_expressions, _ensure_list

    features = _ensure_strings_from_expressions(features)
    feature_names = features or X.get_column_names()

    if y:
        y = _ensure_strings_from_expressions(y)
        y = _ensure_list(y)
        feature_names = [feat for feat in feature_names if feat not in y]

    multioutput = len(y) > 1

    if multioutput:
        for i in range(len(X)):
            yield (
                {key: X.evaluate(key, i, i + 1)[0] for key in feature_names},
                {key: X.evaluate(key, i, i + 1)[0] for key in y}
            )

    else:

        for i in range(len(X)):
            yield (
                {key: X.evaluate(key, i, i + 1)[0] for key in feature_names},
                X.evaluate(y[0], i, i + 1)[0]
            )


class DictReader(csv.DictReader):
    """Overlay on top of `csv.DictReader` which allows sampling."""

    def __init__(self, fraction, rng, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fraction = fraction
        self.rng = rng

    def __next__(self):

        if self.line_num == 0:
            self.fieldnames

        row = next(self.reader)

        if self.fraction < 1:
            while self.rng.random() > self.fraction:
                row = next(self.reader)

        self.line_num = self.reader.line_num

        while row == []:
            row = next(self.reader)
        d = dict(zip(self.fieldnames, row))
        lf = len(self.fieldnames)
        lr = len(row)
        if lf < lr:
            d[self.restkey] = row[lf:]
        elif lf > lr:
            for key in self.fieldnames[lr:]:
                d[key] = self.restval
        return d


def iter_csv(filepath_or_buffer, target_name, names=None, converters=None, parse_dates=None, fraction=1.,
             compression='infer', seed=None):
    """Yields rows from a CSV file.

    Parameters:
        filepath_or_buffer: Either a string indicating the location of a CSV file, or a buffer
            object that has a ``read`` method.
        target_name (str): The name of the target.
        names (list of str): A list of names to associate with each element in a row. If ``None``,
            then the first row will be assumed to contain the names.
        converters (dict): A `dict` mapping feature names to callables used to parse their
            associated values.
        parse_dates (dict): A `dict` mapping feature names to a format passed to the
            `datetime.datetime.strptime` method.
        fraction (float): Sampling fraction.
        compression (str): For on-the-fly decompression of on-disk data. If 'infer' and
            ``filepath_or_buffer`` is path-like, then the decompression method is inferred for the
            following extensions: '.gz'.
        seed (int): If specified, the sampling will be deterministic.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    Example:

        ::

            >>> import io
            >>> from creme import stream

            >>> data = io.StringIO('''name,day,viewers
            ... Breaking Bad,2018-03-14,1337
            ... The Sopranos,2018-03-14,42
            ... Breaking Bad,2018-03-15,7331
            ... ''')

            >>> params = dict(
            ...     target_name='viewers',
            ...     converters={'viewers': int},
            ...     parse_dates={'day': '%Y-%m-%d'}
            ... )

            >>> for x, y in stream.iter_csv(data, **params):
            ...     print(x, y)
            {'name': 'Breaking Bad', 'day': datetime.datetime(2018, 3, 14, 0, 0)} 1337
            {'name': 'The Sopranos', 'day': datetime.datetime(2018, 3, 14, 0, 0)} 42
            {'name': 'Breaking Bad', 'day': datetime.datetime(2018, 3, 15, 0, 0)} 7331

    """

    # If a file is not opened, then we open it
    if not hasattr(filepath_or_buffer, 'read'):

        # Determine the compression from the file extension if "infer" has been specified
        if compression == 'infer':
            _, ext = os.path.splitext(filepath_or_buffer)
            compression = {
                '.csv': 'csv',
                '.gz': 'gzip'
            }[ext]

        # Determine the file opening method from the compression
        open_func = {
            'csv': open,
            'gzip': functools.partial(gzip.open, mode='rt')
        }[compression]

        # Open the file using the opening method
        filepath_or_buffer = open_func(filepath_or_buffer)

    for x in DictReader(
        fraction=fraction,
        rng=random.Random(seed),
        f=filepath_or_buffer,
        fieldnames=names
    ):

        # Cast the values to the given types
        if converters is not None:
            for i, t in converters.items():
                x[i] = t(x[i])

        # Parse the dates
        if parse_dates is not None:
            for i, fmt in parse_dates.items():
                x[i] = dt.datetime.strptime(x[i], fmt)

        # Separate the target from the features
        y = x.pop(target_name)

        yield x, y

    # Close the file
    filepath_or_buffer.close()


def simulate_qa(X_y, on, lag):
    """Simulate a real case learning scenario using a temporal attribute and a lag.

    Each observation will first be shown without revealing the ``y`` value. Once the observation is
    old enough it will be shown once again, this time with the ``y`` value being revealed. Each
    observation will thus be shown twice, once without the target equal to ``None`` and once with
    the actual value. The duration between "questions" and "answers" depends on the ``lag``
    parameter. For this to work, the data is assumed to be sorted with respect to the temporal
    attribute.

    Parameters:
        X_y (generator): A stream of (``x``, ``y``) pairs.
        on (str): The attribute used for measuring time.
        lag (datetime.timedelta or int or float): Amount to wait before revealing the target
            associated with each observation. This value is expected to be able to sum with the
            ``on`` attribute.

    Yields:
        (``q_a``, ``x``, ``y``) pairs where ``q_a`` is a `bool` that indicates if the current
            iteration is question or not.

    Example:

        >>> from creme import stream

        >>> X_y = [({'moment': i, 'x': i}, bool(i % 2)) for i in range(5)]

        >>> for is_question, x, y in simulate_qa(X_y, on='moment', lag=3):
        ...     print('Q' if is_question else 'A', x, y)
        Q {'moment': 0, 'x': 0} False
        Q {'moment': 1, 'x': 1} True
        Q {'moment': 2, 'x': 2} False
        A {'moment': 0, 'x': 0} False
        Q {'moment': 3, 'x': 3} True
        A {'moment': 1, 'x': 1} True
        Q {'moment': 4, 'x': 4} False
        A {'moment': 2, 'x': 2} False
        A {'moment': 3, 'x': 3} True
        A {'moment': 4, 'x': 4} False

    """

    answers = []

    for x, y in X_y:

        while answers:

            # Get the oldest example
            x_ans, y_ans = answers[0]

            # If the oldest answer isn't old enough then stop
            if x_ans[on] + lag > x[on]:
                break

            # Else yield the oldest answer and release it from memory
            yield False, x_ans, y_ans
            del answers[0]

        # Show the observation and label it as a question
        yield True, x, y

        # Store the answer for the future
        answers.append((x, y))

    # Yield the final answers that remain
    for x, y in answers:
        yield False, x, y


def shuffle(stream, buffer_size, seed=None):
    """Shuffles a stream of data.

    This works by maintaining a buffer of elements. The first buffer_size elements are stored in
    memory. Once the buffer is full, a random element inside the buffer is yielded. Everytime an
    element is yielded, the next element in the stream replaces it and the buffer is sampled again.
    Increasing buffer_size will improve the quality of the shuffling.

    If you really want to stream over your dataset in a "good" random order, the best way is to
    split your dataset into smaller datasets and loop over them in a round-robin fashion. You may
    do this by using the ``roundrobin`` recipe from the `itertools` module.

    Example:

        ::

            >>> from creme import stream

            >>> for i in stream.shuffle(range(15), buffer_size=5, seed=42):
            ...     print(i)
            0
            5
            2
            1
            8
            9
            6
            4
            11
            12
            10
            7
            14
            13
            3

    References:
        1. `Visualizing TensorFlow's streaming shufflers <http://www.moderndescartes.com/essays/shuffle_viz/>`_

    """

    rng = random.Random(seed)

    # If stream is not a generator, then we coerce it to one
    if not isinstance(stream, types.GeneratorType):
        stream = iter(stream)

    # Initialize the buffer
    buff = list(itertools.islice(stream, buffer_size))

    # Deplete the stream until it is empty
    for element in stream:

        # Pick a random element from the buffer and yield it
        i = rng.randint(0, len(buff) - 1)
        yield buff[i]

        # Replace the yielded element from the buffer with the new element from the stream
        buff[i] = element

    # Shuffle the remaining buffer elements and yield them one by one
    rng.shuffle(buff)
    for element in buff:
        yield element



cimport base

from . import base
from . import summing


cdef class Mean(base.Univariate):
    """Running mean.

    Parameters:
        mean (float): Initial mean.
        n (float): Initial sum of weights.

    Attributes:
        mean (float): The current value of the mean.
        n (float): The current sum of weights. If each passed weight was 1, then this is equal to
            the number of seen observations.

    Example:

        ::

            >>> from creme import stats

            >>> X = [-5, -3, -1, 1, 3, 5]
            >>> mean = stats.Mean()
            >>> for x in X:
            ...     print(mean.update(x).get())
            -5.0
            -4.0
            -3.0
            -2.0
            -1.0
            0.0

    References:
        1. `Incremental calculation of weighted mean and variance <https://fanf2.user.srcf.net/hermes/doc/antiforgery/stats.pdf>`_

    """

    def __init__(self, mean=0., n=0.):
        self.mean = mean
        self.n = n

    cpdef Mean update(self, double x, double w=1.):
        self.n += w
        self.mean += w * (x - self.mean) / self.n
        return self

    cpdef Mean revert(self, double x, double w=1.):
        self.n -= w
        if self.n < 0:
            raise ValueError('Cannot go below 0')
        elif self.n == 0:
            self.mean = 0.
        else:
            self.mean -= w * (x - self.mean) / self.n
        return self

    cpdef double get(self):
        return self.mean


class RollingMean(summing.RollingSum):
    """Running average over a window.

    Parameters:
        window_size (int): Size of the rolling window.

    Example:

        ::

            >>> import creme

            >>> X = [1, 2, 3, 4, 5, 6]

            >>> rolling_mean = creme.stats.RollingMean(window_size=2)
            >>> for x in X:
            ...     print(rolling_mean.update(x).get())
            1.0
            1.5
            2.5
            3.5
            4.5
            5.5

            >>> rolling_mean = creme.stats.RollingMean(window_size=3)
            >>> for x in X:
            ...     print(rolling_mean.update(x).get())
            1.0
            1.5
            2.0
            3.0
            4.0
            5.0

    """

    def get(self):
        return super().get() / len(self) if len(self) > 0 else 0


class BayesianMean(base.Univariate):
    """Estimates a mean using outside information.

    References:

        1. `Additive smoothing <https://www.wikiwand.com/en/Additive_smoothing>`_
        2. `Bayesian average <https://www.wikiwand.com/en/Bayesian_average>`_
        3. `Practical example of Bayes estimators <https://www.wikiwand.com/en/Bayes_estimator#/Practical_example_of_Bayes_estimators>`_

    """

    def __init__(self, prior, prior_weight):
        self.prior = prior
        self.prior_weight = prior_weight
        self.mean = Mean()

    @property
    def name(self):
        return 'bayes_mean'

    def update(self, x):
        self.mean.update(x)
        return self

    def get(self):

        # Uses the notation from https://www.wikiwand.com/en/Bayes_estimator#/Practical_example_of_Bayes_estimators
        R = self.mean.get()
        v = self.mean.n
        m = self.prior_weight
        C = self.prior

        return (R * v + C * m) / (v + m)



import collections

import numpy as np

from . import base


__all__ = ['FTRLProximal']


class FTRLProximal(base.Optimizer):
    """FTRL-Proximal optimizer.

    Example:

        ::

            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_breast_cancer(),
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> optimizer = optim.FTRLProximal()
            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     linear_model.LogisticRegression(optimizer)
            ... )
            >>> metric = metrics.F1()

            >>> model_selection.online_score(X_y, model, metric)
            F1: 0.962132

    References:
        1. `Ad Click Prediction: a View from the Trenches - Research - Google <https://static.googleusercontent.com/media/research.google.com/en//pubs/archive/41159.pdf>`_
        2. `tf.train.FtrlOptimizer <https://www.tensorflow.org/api_docs/python/tf/train/FtrlOptimizer>`_

    """

    def __init__(self, alpha=.05, beta=1., l1=0., l2=1.):
        self.alpha = alpha
        self.beta = beta
        self.l1 = l1
        self.l2 = l2
        self.z = collections.defaultdict(float)
        self.n = collections.defaultdict(float)
        self.n_iterations = 0

    def _update_after_pred(self, w, g):

        alpha = self.alpha
        beta = self.beta
        l1 = self.l1
        l2 = self.l2
        z = self.z
        n = self.n

        for i in g:
            if abs(z[i]) > l1:
                w[i] = -((beta + n[i] ** 0.5) / alpha + l2) ** -1 * (z[i] - np.sign(z[i]) * l1)

        for i, gi in g.items():
            s = ((self.n[i] + gi ** 2) ** 0.5 - self.n[i] ** 0.5) / self.alpha
            self.z[i] += gi - s * w.get(i, 0)
            self.n[i] += gi ** 2

        return w



from .. import base


__all__ = ['OneHotEncoder']


class OneHotEncoder(base.Transformer):
    """One-hot encoding.

    Attributes:
        on (str): Attribute to one-hot encode.
        sparse (bool): Whether or not 0s should be made explicit or not.

    Example:

        ::

            >>> import pprint
            >>> import string
            >>> import creme.preprocessing
            >>> import numpy as np

            >>> rng = np.random.RandomState(42)
            >>> alphabet = list(string.ascii_lowercase)
            >>> X = [{'letter': letter} for letter in rng.choice(alphabet, size=10)]

            >>> one_hot = creme.preprocessing.OneHotEncoder('letter', sparse=True)
            >>> for x in X:
            ...     print(one_hot.fit_one(x).transform_one(x))
            {'letter_g': 1}
            {'letter_t': 1}
            {'letter_o': 1}
            {'letter_k': 1}
            {'letter_h': 1}
            {'letter_u': 1}
            {'letter_g': 1}
            {'letter_z': 1}
            {'letter_s': 1}
            {'letter_w': 1}

            >>> one_hot = creme.preprocessing.OneHotEncoder('letter')
            >>> for letter in ['a', 'b', 'c']:
            ...     x = {'letter': letter}
            ...     pprint.pprint(one_hot.fit_one(x).transform_one(x))
            {'letter_a': 1}
            {'letter_a': 0, 'letter_b': 1}
            {'letter_a': 0, 'letter_b': 0, 'letter_c': 1}

    """

    def __init__(self, on, sparse=False):
        self.on = on
        self.sparse = sparse
        self.values = set()

    def fit_one(self, x, y=None):
        self.values.add(x[self.on])
        return self

    def transform_one(self, x, y=None):
        oh = {}
        if not self.sparse:
            oh = {f'{self.on}_{i}': 0 for i in self.values}
        oh[f'{self.on}_{x[self.on]}'] = 1
        return oh

    def __str__(self):
        return f'OneHotEncoder({self.on})'



import abc

from .. import utils


cdef class Statistic:

    cpdef double get(self):
        """Returns the current value of the statistic."""
        raise NotImplementedError

    def __str__(self):
        return f'{self.__class__.__name__}: {self.get():.6f}'.rstrip('0')

    def __repr__(self):
        return str(self)


cdef class Univariate(Statistic):

    cpdef Univariate update(self, double x):
        """Updates and returns the called instance."""
        raise NotImplementedError

    @property
    def name(self):
        return self.__class__.__name__.lower()


class RollingUnivariate(Univariate):

    @abc.abstractproperty
    def window_size(self):
        pass

    @property
    def name(self):
        return f'rolling_{super().name}_{self.window_size}'


cdef class Bivariate(Statistic):

    cpdef Bivariate update(self, double x, double y):
        """Updates and returns the called instance."""
        raise NotImplementedError



import collections
import operator

from .. import base
from .. import utils


__all__ = ['KNeighborsRegressor', 'KNeighborsClassifier']


class NearestNeighbours(collections.deque):

    def __init__(self, window_size, p):
        super().__init__(self, maxlen=window_size)
        self.p = p

    def update(self, x, y):
        super().append((x, y))
        return self

    def find_nearest(self, x, k):
        """Returns the ``k`` closest points, along with their distances."""

        # Compute the distances to each point in the window
        points = ((*p, utils.math.minkowski_distance(a=x, b=p[0], p=self.p)) for p in self)

        # Return the k closest points
        return sorted(points, key=operator.itemgetter(2))[:k]


class KNeighborsRegressor(NearestNeighbours, base.Regressor):
    """K-Nearest Neighbors (KNN) for regression.

    Parameters:
        n_neighbors (int): Number of neighbors to use.
        window_size (int): Size of the sliding window use to search neighbors with.
        p (int): Power parameter for the Minkowski metric. When ``p=1``, this corresponds to the
            Manhattan distance, while ``p=2`` corresponds to the Euclidean distance.
        weighted (bool): Whether to weight the contribution of each neighbor by it's inverse
            distance or not.

    Example:

        ::

            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import neighbors
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_boston(),
            ...     shuffle=True,
            ...     random_state=42
            ... )

            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     neighbors.KNeighborsRegressor()
            ... )

            >>> metric = metrics.MAE()

            >>> model_selection.online_score(X_y, model, metric)
            MAE: 3.817195

    """

    def __init__(self, n_neighbors=5, window_size=50, p=2, weighted=True):
        super().__init__(window_size=window_size, p=p)
        self.n_neighbors = n_neighbors
        self.weighted = weighted

    def fit_one(self, x, y):
        return super().update(x, y)

    def predict_one(self, x):

        nearest = self.find_nearest(x=x, k=self.n_neighbors)

        if not nearest:
            return 0.

        # If the closest neighbor has a distance of 0, then return it's output
        if nearest[0][2] == 0:
            return nearest[0][1]

        # Weighted average
        if self.weighted:
            return (
                sum(y / d for _, y, d in nearest) /
                sum(1 / d for _, _, d in nearest)
            )

        # Uniform average
        return sum(y for _, y, _ in nearest) / self.n_neighbors


class KNeighborsClassifier(NearestNeighbours, base.MultiClassifier):
    """K-Nearest Neighbors (KNN) for classification.

    Parameters:
        n_neighbors (int): Number of neighbors to use.
        window_size (int): Size of the sliding window use to search neighbors with.
        p (int): Power parameter for the Minkowski metric. When ``p=1``, this corresponds to the
            Manhattan distance, while ``p=2`` corresponds to the Euclidean distance.
        weighted (bool): Whether to weight the contribution of each neighbor by it's inverse
            distance or not.

    Example:

        ::

            >>> from creme import datasets
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import neighbors
            >>> from creme import preprocessing

            >>> X_y = datasets.fetch_electricity()

            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     neighbors.KNeighborsClassifier()
            ... )

            >>> metric = metrics.Accuracy()

            >>> model_selection.online_score(X_y, model, metric)
            Accuracy: 0.88526

    """

    def __init__(self, n_neighbors=5, window_size=50, p=2, weighted=True):
        super().__init__(window_size=window_size, p=p)
        self.n_neighbors = n_neighbors
        self.weighted = weighted
        self.classes = set()

    def fit_one(self, x, y):
        self.classes.add(y)
        return super().update(x, y)

    def predict_proba_one(self, x):

        nearest = self.find_nearest(x=x, k=self.n_neighbors)

        y_pred = {c: 0. for c in self.classes}

        if not nearest:
            return y_pred

        # If the closest neighbor has a distance of 0, then return it's output
        if nearest[0][2] == 0:
            y_pred[nearest[0][1]] = 1.
            return y_pred

        # Weighted votes
        if self.weighted:
            for _, y, d in nearest:
                y_pred[y] += 1. / d

        # Uniform votes
        else:
            for _, y, _ in nearest:
                y_pred[y] += 1.

        # Normalize votes into real [0, 1] probabilities
        return utils.math.softmax(y_pred)



import math

from .. import stats

from . import base


__all__ = ['LogLoss']


class BaseLogLoss(base.BinaryMetric):

    @property
    def bigger_is_better(self):
        return False

    @property
    def requires_labels(self):
        return False


class LogLoss(stats.Mean, BaseLogLoss):
    """Binary logarithmic loss.

    Example:

        ::

            >>> from creme import metrics
            >>> from sklearn.metrics import log_loss

            >>> y_true = [True, False, False, True]
            >>> y_pred = [0.9,  0.1,   0.2,   0.65]

            >>> metric = metrics.LogLoss()
            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     metric = metric.update(y_t, y_p)
            ...     print(metric.get())
            0.105360...
            0.105360...
            0.144621...
            0.216161...

            >>> metric
            LogLoss: 0.216162

    """

    def _get_log_loss(self, y_true, y_pred):
        p_true = y_pred.get(True, 0.) if isinstance(y_pred, dict) else y_pred
        p_true = self.clamp_proba(p_true)
        if y_true:
            return -math.log(p_true)
        return -math.log(1 - p_true)

    def update(self, y_true, y_pred, sample_weight=1.):
        ll = self._get_log_loss(y_true, y_pred)
        return super().update(x=ll, w=sample_weight)

    def revert(self, y_true, y_pred, sample_weight=1.):
        ll = self._get_log_loss(y_true, y_pred)
        return super().revert(x=ll, w=sample_weight)



from .. import base


__all__ = ['Renamer']


class Renamer(base.Transformer):
    """Renames keys based on given parameters.

    Parameters:
        whitelist (strs): Key(s) to keep.

    Example:

        ::

            >>> from creme import compose

            >>> x = {'a': 42, 'b': 12}
            >>> compose.Renamer(prefix='prefix_', suffix='_suffix').transform_one(x)
            {'prefix_a_suffix': 42, 'prefix_b_suffix': 12}

    """

    def __init__(self, prefix='', suffix=''):
        self.prefix = prefix
        self.suffix = suffix

    def _rename(self, s):
        return self.prefix + s + self.suffix

    def transform_one(self, x):
        return {self._rename(i): xi for i, xi in x.items()}



import abc


class Recommender(abc.ABC):

    @abc.abstractmethod
    def fit_one(self, row_id, col_id, value):
        pass

    @abc.abstractmethod
    def predict_one(self, row_id, col_id):
        pass



"""Feature preprocessing."""
from .feature_hasher import FeatureHasher
from .one_hot import OneHotEncoder
from .poly import PolynomialExtender
from .scale import MinMaxScaler
from .scale import Normalizer
from .scale import StandardScaler


__all__ = [
    'FeatureHasher',
    'MinMaxScaler',
    'Normalizer',
    'OneHotEncoder',
    'PolynomialExtender',
    'StandardScaler'
]



"""Model evaluation and selection."""
import math

from . import base
from . import stream


__all__ = ['online_score', 'online_qa_score']


def online_score(X_y, model, metric, print_every=math.inf):
    """Computes the online score of a model given a stream of data and a metric.

    Parameters:
        X_y (generator): A stream of (features, target) tuples.
        model (base.Estimator)
        metric (metrics.Metric)
        print_every (int): Iteration number at which to print the current metric.

    Returns:
        metric

    References:
        1. `Beating the Hold-Out: Bounds for K-fold and Progressive Cross-Validation <http://hunch.net/~jl/projects/prediction_bounds/progressive_validation/coltfinal.pdf>`_

    """

    # Check that the model and the metric are in accordance
    if not metric.works_with(model):
        raise ValueError(f"{metric.__class__.__name__} metric can't be used to evaluate a " +
                         f'{model.__class__.__name__}')

    # Determine if predict_one or predict_proba_one should be used in case of a classifier
    pred_func = model.predict_one
    if isinstance(model, base.Classifier) and not metric.requires_labels:
        pred_func = model.predict_proba_one

    # Handle the first observation separately
    x, y = next(X_y)
    y_pred = pred_func(x)
    model = model.fit_one(x, y)
    if y_pred != {} and y_pred is not None:
        metric.update(y, y_pred)

    # Train the model and use the out-of-fold predictions to update the metric
    for i, (x, y) in enumerate(X_y):
        y_pred = pred_func(x)
        model = model.fit_one(x, y)
        metric.update(y, y_pred)

        if not (i + 1) % print_every:
            print(f'[{i+1:,d}] {metric}')

    return metric


def online_qa_score(X_y, model, metric, on, lag, print_every=math.inf):
    """A variant of online scoring where the targets are revealed with a lag.

    ``X_y`` is converted into a question and answer where the model is asked to predict an
    observation. The target is only revealed to the model after a certain amount of ``lag``. See
    `stream.simulate_qa` for more information.

    Parameters:
        X_y (generator): A stream of (features, target) tuples.
        model (base.Estimator)
        metric (metrics.Metric)
        on (str): The attribute used for measuring time.
        lag (datetime.timedelta or int or float): Amount to wait before revealing the target
            associated with each observation. This value is expected to be able to sum with the
            ``on`` attribute.
        print_every (int): Iteration number at which to print the current metric. This only takes
            into account the predictions, and not the training steps.

    Returns:
        metric

    """

    # Check that the model and the metric are in accordance
    if not metric.works_with(model):
        raise ValueError(f"{metric.__class__.__name__} metric can't be used to evaluate a " +
                         f'{model.__class__.__name__}')

    # Determine if predict_one or predict_proba_one should be used in case of a classifier
    pred_func = model.predict_one
    if isinstance(model, base.Classifier) and not metric.requires_labels:
        pred_func = model.predict_proba_one

    n_questions = 0

    # Train the model and use the out-of-fold predictions to update the metric
    for is_question, x, y in stream.simulate_qa(X_y, on=on, lag=lag):

        if is_question:
            y_pred = pred_func(x)
            if y_pred != {} and y_pred is not None:
                metric.update(y, y_pred)

            n_questions += 1
            if not n_questions % print_every:
                print(f'[{n_questions:,d}] {metric}')

        else:
            model = model.fit_one(x, y)

    return metric



"""
creme is a library for incremental learning. Incremental learning is a machine learning regime
where the observations are made available one by one. It is also known as online learning,
iterative learning, or sequential learning. This is in contrast to batch learning where all the
data is processed at once. Incremental learning is desirable when the data is too big to fit in
memory, or simply when it isn't available all at once. creme's API is heavily inspired from that of
scikit-learn, enough so that users who are familiar with scikit-learn should feel right at home.
"""
from .__version__ import __version__


__all__ = [
    'anomaly',
    'base',
    'cluster',
    'compat',
    'compose',
    'datasets',
    'dummy',
    'decomposition',
    'ensemble',
    'feature_extraction',
    'feature_selection',
    'impute',
    'linear_model',
    'metrics',
    'model_selection',
    'multiclass',
    'multioutput',
    'naive_bayes',
    'neighbors',
    'optim',
    'preprocessing',
    'proba',
    'reco',
    'stats',
    'stream',
    'time_series',
    'tree',
    'utils'
]



import collections

from . import base
from . import pearson


class AutoCorrelation(base.Univariate):
    """Measures the serial correlation.

    This method computes the Pearson correlation between the current value and the value seen ``n``
    steps before.

    Example:

        The following examples are taken from the `pandas documentation <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.autocorr.html>`_.

        ::

            >>> from creme import stats

            >>> auto_corr = stats.AutoCorrelation(lag=1)
            >>> for x in [0.25, 0.5, 0.2, -0.05]:
            ...     print(auto_corr.update(x).get())
            0
            0
            -1.0
            0.103552...

            >>> auto_corr = stats.AutoCorrelation(lag=2)
            >>> for x in [0.25, 0.5, 0.2, -0.05]:
            ...     print(auto_corr.update(x).get())
            0
            0
            0
            -1.0

            >>> auto_corr = stats.AutoCorrelation(lag=1)
            >>> for x in [1, 0, 0, 0]:
            ...     print(auto_corr.update(x).get())
            0
            0
            0
            0

    """

    def __init__(self, lag):
        self.window = collections.deque(maxlen=lag)
        self.lag = lag
        self.pearson = pearson.PearsonCorrelation(ddof=1)

    @property
    def name(self):
        return f'autocorr_{self.lag}'

    def update(self, x):

        # The correlation can be update once enough elements have been seen
        if len(self.window) == self.lag:
            self.pearson.update(x, self.window[0])

        # Add x to the window
        self.window.append(x)

        return self

    def get(self):
        return self.pearson.get()



"""Utilities for unit testing and sanity checking estimators."""
import copy
import functools
import itertools
import math
import pickle
import random

from sklearn import datasets


__all__ = ['check_estimator', 'guess_model']


def guess_model(model):

    from .. import base
    from .. import compose

    if isinstance(model, compose.Pipeline):
        return guess_model(model.final_estimator)
    elif isinstance(model, base.Wrapper):
        return guess_model(model._model)
    return model


def make_random_sentence():

    nouns = ('puppy', 'car', 'rabbit', 'girl', 'monkey')
    verbs = ('runs', 'hits', 'jumps', 'drives', 'barfs')
    adj = ('adorable', 'clueless', 'dirty', 'odd', 'stupid')
    adv = ('crazily', 'dutifully', 'foolishly', 'merrily', 'occasionally')

    return ' '.join([random.choice(i) for i in [nouns, verbs, adj, adv]])


def make_random_features(model, n_observations, n_features):

    model = guess_model(model)
    tags = model._get_tags()

    if tags['handles_text']:
        for _ in range(n_observations):
            yield make_random_sentence()

    elif tags['requires_positive_data']:
        for _ in range(n_observations):
            yield {i: random.uniform(0, 10) for i in range(n_features)}

    else:
        for _ in range(n_observations):
            yield {i: random.random() for i in range(n_features)}


def make_random_targets(model, n_observations):

    from .. import base

    random_func = None
    model = guess_model(model)

    if isinstance(model, base.Regressor):
        random_func = random.random

    elif isinstance(model, base.Ensemble):
        if any(isinstance(m, base.BinaryClassifier) for m in model):
            random_func = functools.partial(random.choice, [True, False])
        else:
            random_func = functools.partial(random.choice, ['a', 'b', 'c', 'd'])

    elif isinstance(model, base.MultiClassifier):
        random_func = functools.partial(random.choice, ['a', 'b', 'c', 'd'])

    elif isinstance(model, base.BinaryClassifier):
        random_func = functools.partial(random.choice, [True, False])

    elif isinstance(model, base.Transformer):
        if model.is_supervised:
            random_func = random.random
        else:
            random_func = lambda: None

    elif isinstance(model, base.Clusterer):
        random_func = functools.partial(random.choice, [0, 1, 2, 3])

    for _ in range(n_observations):
        yield random_func()


def make_random_X_y(model, n_observations, n_features):
    return zip(
        make_random_features(model, n_observations, n_features),
        make_random_targets(model, n_observations)
    )


def check_fit_one(model):

    klass = model.__class__

    for x, y in make_random_X_y(model, n_observations=300, n_features=8):

        xx, yy = copy.deepcopy(x), copy.deepcopy(y)

        model = model.fit_one(x, y)

        # Check the model returns itself
        assert isinstance(model, klass)

        # Check fit_one is pure (i.e. x and y haven't changed)
        assert x == xx
        assert y == yy


def check_predict_proba_one(model):

    for x, y in make_random_X_y(model, n_observations=300, n_features=8):

        xx, yy = copy.deepcopy(x), copy.deepcopy(y)

        model = model.fit_one(x, y)
        y_pred = model.predict_proba_one(x)

        # Check the probabilities are coherent
        assert isinstance(y_pred, dict)
        assert math.isclose(sum(y_pred.values()), 1.)
        for proba in y_pred.values():
            assert 0. <= proba <= 1.

        # Check predict_proba_one is pure (i.e. x and y haven't changed)
        assert x == xx
        assert y == yy


def check_a_better_than_b(model_a, model_b, X_y_func, metric):
    """Trains two models and checks that model_a does better than model_b."""

    from .. import model_selection

    metric_a = model_selection.online_score(
        X_y=X_y_func(),
        model=model_a,
        metric=copy.deepcopy(metric)
    )
    metric_b = model_selection.online_score(
        X_y=X_y_func(),
        model=model_b,
        metric=copy.deepcopy(metric)
    )

    if metric.bigger_is_better:
        assert metric_a.get() > metric_b.get()
    else:
        assert metric_a.get() < metric_b.get()


def check_better_than_dummy_binary(classifier):

    from .. import dummy
    from .. import metrics
    from .. import stream

    for dummy_model, X_y_func, metric in itertools.product(
        (dummy.NoChangeClassifier(), dummy.PriorClassifier()),
        (functools.partial(
            stream.iter_sklearn_dataset,
            datasets.load_breast_cancer(),
            shuffle=True,
            random_state=42,
        ),),
        (metrics.Accuracy(),)
    ):
        check_a_better_than_b(
            model_a=copy.deepcopy(classifier),
            model_b=copy.deepcopy(dummy_model),
            X_y_func=X_y_func,
            metric=copy.deepcopy(metric)
        )


def check_better_than_dummy_multi(classifier):

    from .. import dummy
    from .. import metrics
    from .. import stream

    for dummy_model, X_y_func, metric in itertools.product(
        (dummy.NoChangeClassifier(), dummy.PriorClassifier()),
        (functools.partial(
            stream.iter_sklearn_dataset,
            datasets.load_iris(),
            shuffle=True,
            random_state=42,
        ),),
        (metrics.Accuracy(),)
    ):
        check_a_better_than_b(
            model_a=copy.deepcopy(classifier),
            model_b=copy.deepcopy(dummy_model),
            X_y_func=X_y_func,
            metric=copy.deepcopy(metric)
        )


def check_better_than_dummy_regression(regressor):

    from .. import dummy
    from .. import metrics
    from .. import stats
    from .. import stream

    for dummy_model, X_y_func, metric in itertools.product(
        (dummy.StatisticRegressor(stats.Mean()),),
        (functools.partial(
            stream.iter_sklearn_dataset,
            datasets.load_boston(),
            shuffle=True,
            random_state=42,
        ),),
        (metrics.MSE(),)
    ):
        check_a_better_than_b(
            model_a=copy.deepcopy(regressor),
            model_b=copy.deepcopy(dummy_model),
            X_y_func=X_y_func,
            metric=copy.deepcopy(metric)
        )


def check_predict_proba_one_binary(classifier):

    for x, y in make_random_X_y(classifier, n_observations=300, n_features=8):
        y_pred = classifier.predict_proba_one(x)
        classifier = classifier.fit_one(x, y)
        assert len(y_pred) == 2
        assert True in y_pred
        assert False in y_pred


def check_pickling(model):
    assert isinstance(pickle.loads(pickle.dumps(model)), model.__class__)


def check_repr(model):
    assert isinstance(repr(model), str)


def check_str(model):
    assert isinstance(str(model), str)


def check_debug_one(model):
    for x, y in make_random_X_y(model, n_observations=10, n_features=4):
        model.fit_one(x, y)
    model.debug_one(x)


def yield_all_checks(model):

    from .. import base

    if isinstance(model, base.OutlierDetector):
        return
        yield

    yield check_fit_one
    yield check_pickling
    yield check_repr
    yield check_str
    if hasattr(model, 'debug_one'):
        yield check_debug_one

    model = guess_model(model)
    tags = model._get_tags()

    if isinstance(model, base.Classifier):
        yield check_predict_proba_one

        if not isinstance(model, base.MultiClassifier):
            yield check_predict_proba_one_binary

    # if not tags['poor_score']:
    #     if isinstance(model, base.BinaryClassifier):
    #         yield check_better_than_dummy_binary
    #     if isinstance(model, base.MultiClassifier):
    #         yield check_better_than_dummy_binary
    #         yield check_better_than_dummy_multi
    #     if isinstance(model, base.Regressor):
    #         yield check_better_than_dummy_regression


def check_estimator(model):
    """Check if a model adheres to ``creme``'s API.

    Parameters:
        model (base.Estimator)

    """

    for check in yield_all_checks(model):
        check(copy.deepcopy(model))



import collections
import copy
import functools

from . import base


__all__ = ['PerClass']


class PerClass(base.WrapperMetric, base.MultiClassMetric):
    """Wrapper for computing metrics per class.

    Parameters:
        metric (metrics.BinaryMetric)

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = [0, 0, 1, 1, 2, 1, 0]
            >>> y_pred = [0, 0, 0, 1, 1, 1, 1]

            >>> metric = metrics.PerClass(metrics.Accuracy())

            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     print(metric.update(y_t, y_p))
            ...     print('---')
            0: Accuracy: 1.
            ---
            0: Accuracy: 1.
            ---
            0: Accuracy: 1.
            1: Accuracy: 0.
            ---
            0: Accuracy: 1.
            1: Accuracy: 0.5
            ---
            0: Accuracy: 1.
            1: Accuracy: 0.5
            2: Accuracy: 0.
            ---
            0: Accuracy: 1.
            1: Accuracy: 0.666667
            2: Accuracy: 0.
            ---
            0: Accuracy: 0.666667
            1: Accuracy: 0.666667
            2: Accuracy: 0.
            ---

        You can combine this with `metrics.Rolling` to obtain rolling metrics for each class.

        ::

            >>> from creme import metrics

            >>> y_true = [0, 0, 1, 1, 2, 1, 0]
            >>> y_pred = [0, 0, 0, 1, 1, 1, 1]

            >>> metric = metrics.Rolling(metrics.PerClass(metrics.Accuracy()), window_size=2)

            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     print(metric.update(y_t, y_p))
            ...     print('---')
            Rolling of size 2
                0: Accuracy: 1.
            ---
            Rolling of size 2
                0: Accuracy: 1.
            ---
            Rolling of size 2
                0: Accuracy: 1.
                1: Accuracy: 0.
            ---
            Rolling of size 2
                0: Accuracy: 0.
                1: Accuracy: 0.5
            ---
            Rolling of size 2
                1: Accuracy: 1.
                2: Accuracy: 0.
            ---
            Rolling of size 2
                1: Accuracy: 1.
                2: Accuracy: 0.
            ---
            Rolling of size 2
                0: Accuracy: 0.
                1: Accuracy: 1.
            ---

    Note:
        Note that ``Rolling(PerClass(metric))`` will produce different results than
        ``PerClass(Rolling(metric))``.

    """

    def __init__(self, metric):
        self.metrics = collections.defaultdict(functools.partial(copy.deepcopy, metric))
        self._metric = metric
        self._class_counts = collections.Counter()

    @property
    def metric(self):
        return self._metric

    @property
    def requires_labels(self):
        return self._metric.requires_labels

    def update(self, y_true, y_pred, sample_weight=1.):
        self.metrics[y_true].update(y_true, y_pred, sample_weight)
        self._class_counts.update([y_true, y_pred])
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self.metrics[y_true].revert(y_true, y_pred, sample_weight)
        self._class_counts.subtract([y_true, y_pred])
        return self

    def get(self):
        return {c: m.get() for c, m in self.metrics.items() if self._class_counts[c] > 0}

    def __str__(self):
        return '\n'.join(
            f'{c}: {str(m)}'
            for c, m in self.metrics.items()
            if self._class_counts[c] > 0
        )



import collections
import copy
import statistics

from sklearn import utils

from .. import base


__all__ = ['BaggingClassifier', 'BaggingRegressor']


class BaseBagging(base.Wrapper, base.Ensemble):

    def __init__(self, model, n_models=10, random_state=None):
        super().__init__(copy.deepcopy(model) for i in range(n_models))
        self.model = model
        self.rng = utils.check_random_state(random_state)

    @property
    def _model(self):
        return self.model

    def fit_one(self, x, y):

        for model in self:
            for _ in range(self.rng.poisson(1)):
                model.fit_one(x, y)

        return self


class BaggingClassifier(BaseBagging, base.Classifier):
    """Online bootstrap aggregation for classification.

    For each incoming observation, each model's ``fit_one`` method is called ``k`` times where
    ``k`` is sampled from a Poisson distribution of parameter 1. ``k`` thus has a 36% chance of
    being equal to 0, a 36% chance of being equal to 1, an 18% chance of being equal to 2, a 6%
    chance of being equal to 3, a 1% chance of being equal to 4, etc. You can do
    ``scipy.stats.poisson(1).pmf(k)`` to obtain more detailed values.

    Parameters:
        model (BinaryClassifier or MultiClassifier): The classifier to bag.
        n_models (int): The number of models in the ensemble.
        random_state (int, ``numpy.random.RandomState`` instance or None): If int, ``random_state``
            is the seed used by the random number generator; if ``RandomState`` instance,
            ``random_state`` is the random number generator; if ``None``, the random number
            generator is the ``RandomState`` instance used by `numpy.random`.

    Example:

        In the following example three logistic regressions are bagged together. The performance is
        slightly better than when using a single logistic regression.

        ::

            >>> from creme import compose
            >>> from creme import ensemble
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_breast_cancer(),
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> model = ensemble.BaggingClassifier(
            ...     model=(
            ...         preprocessing.StandardScaler() |
            ...         linear_model.LogisticRegression()
            ...     ),
            ...     n_models=3,
            ...     random_state=42
            ... )
            >>> metric = metrics.F1()

            >>> model_selection.online_score(X_y, model, metric)
            F1: 0.963889

            >>> print(model)
            BaggingClassifier(StandardScaler | LogisticRegression)

    References:
        1. `Online Bagging and Boosting <https://ti.arc.nasa.gov/m/profile/oza/files/ozru01a.pdf>`_

    """

    def predict_proba_one(self, x):
        """Averages the predictions of each classifier."""

        y_pred = collections.Counter()
        for classifier in self:
            y_pred.update(classifier.predict_proba_one(x))

        total = sum(y_pred.values())
        return {label: proba / total for label, proba in y_pred.items()}


class BaggingRegressor(BaseBagging, base.Regressor):
    """Online bootstrap aggregation for regression.

    For each incoming observation, each model's ``fit_one`` method is called ``k`` times where
    ``k`` is sampled from a Poisson distribution of parameter 1. ``k`` thus has a 36% chance of
    being equal to 0, a 36% chance of being equal to 1, an 18% chance of being equal to 2, a 6%
    chance of being equal to 3, a 1% chance of being equal to 4, etc. You can do
    ``scipy.stats.poisson(1).pmf(k)`` for more detailed values.

    Parameters:
        model (Regressor): The regressor to bag.
        n_models (int): The number of models in the ensemble.
        random_state (int, ``numpy.random.RandomState`` instance or None): If int, ``random_state``
            is the seed used by the random number generator; if ``RandomState`` instance,
            ``random_state`` is the random number generator; if ``None``, the random number
            generator is the ``RandomState`` instance used by `numpy.random`.

    Example:

        In the following example three logistic regressions are bagged together. The performance is
        slightly better than when using a single logistic regression.

        ::

            >>> from creme import compose
            >>> from creme import ensemble
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_boston(),
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> model = preprocessing.StandardScaler()
            >>> model |= ensemble.BaggingRegressor(
            ...     model=linear_model.LinearRegression(intercept_lr=0.1),
            ...     n_models=3,
            ...     random_state=42
            ... )
            >>> metric = metrics.MAE()

            >>> model_selection.online_score(X_y, model, metric)
            MAE: 4.26101

    References:
        1. `Online Bagging and Boosting <https://ti.arc.nasa.gov/m/profile/oza/files/ozru01a.pdf>`_

    """

    def predict_one(self, x):
        """Averages the predictions of each regressor."""
        return statistics.mean((regressor.predict_one(x) for regressor in self))



"""Weight initialization schemes."""
import numpy as np
from sklearn import utils


__all__ = [
    'Constant',
    'Normal',
    'Zeros'
]


class Constant:
    """Constant initializer which always return the same value.

    Parameters:
        value (float): The constant value

    Example:

        ::

            >>> from creme import optim

            >>> init = optim.initializers.Constant(value=3.14)

            >>> init(shape=1)
            3.14

            >>> init(shape=2)
            array([3.14, 3.14])

    """

    def __init__(self, value):
        self.value = value

    def __call__(self, shape):
        return np.full(shape, self.value) if shape != 1 else self.value


class Zeros(Constant):
    """Initializer which return zeros for each new weight.

    Example:

        ::

            >>> from creme import optim

            >>> init = optim.initializers.Zeros()

            >>> init(shape=1)
            0.0

            >>> init(shape=2)
            array([0., 0.])

    """

    def __init__(self):
        super().__init__(value=0.)


class Normal:
    """Random normal initializer which simulate a normal distribution with specified parameters.

    Parameters:
        mu (float): The mean of the normal distribution
        sigma (float): The standard deviation of the normal distribution

    Example:

        ::

            >>> from creme import optim

            >>> init = optim.initializers.Normal(mu=0, sigma=1, random_state=42)

            >>> init(shape=1)
            0.496714...

            >>> init(shape=2)
            array([-0.1382643 ,  0.64768854])

    """

    def __init__(self, mu=0., sigma=1., random_state=None):
        self.mu = mu
        self.sigma = sigma
        self.random_state = utils.check_random_state(random_state)

    def __call__(self, shape):
        weights = self.random_state.normal(loc=self.mu, scale=self.sigma, size=shape)
        if shape == 1:
            return weights[0]
        return weights



import collections
import functools

from .. import base
from .. import stats


class Detrender(base.Regressor, base.Wrapper):
    """A linear detrender which centers the target in zero.

    At each ``fit_one`` step, the current mean of ``y`` is substracted from ``y`` before being fed
    to the provided regression model. During the ``predict_one`` step, the current mean is added
    to the prediction of the regression model.

    Parameters:
        regressor (base.Regressor)
        window_size (int): Window size used for calculating the rolling mean. If ``None``, then a
            mean over the whole target data will instead be used.

    """

    def __init__(self, regressor, window_size=None):
        self.regressor = regressor
        self.mean = stats.Mean() if window_size is None else stats.RollingMean(window_size)

    @property
    def _model(self):
        return self.regressor

    def fit_one(self, x, y):
        self.regressor.fit_one(x, y - self.mean.get())
        self.mean.update(y)
        return self

    def predict_one(self, x):
        return self.regressor.predict_one(x) + self.mean.get()


class GroupDetrender(base.Regressor, base.Wrapper):
    """Removes the trend of the target inside each group.

    Parameters:
        regressor (base.Regressor)
        by (str)
        window_size (int): Window size used for calculating each rolling mean. If ``None``, then a
            mean over the whole target data will instead be used.

    """

    def __init__(self, regressor, by, window_size=None):
        self.regressor = regressor
        self.by = by
        self.means = collections.defaultdict(
            stats.Mean if window_size is None else
            functools.partial(stats.RollingMean, window_size)
        )

    @property
    def _model(self):
        return self.regressor

    def fit_one(self, x, y):
        key = x[self.by]
        self.regressor.fit_one(x, y - self.means[key].get())
        self.means[key].update(y)
        return self

    def predict_one(self, x):
        key = x[self.by]
        return self.regressor.predict_one(x) + self.means[key].get()



"""Toy datasets."""
import ast
import os
import shutil
import tarfile
import urllib
import zipfile

from .. import stream


__all__ = [
    'fetch_bikes',
    'fetch_credit_card',
    'fetch_electricity',
    'fetch_kdd99_http',
    'fetch_restaurants',
    'fetch_sms',
    'fetch_trec07p',
    'load_airline',
    'load_chick_weights'
]


def get_data_home(data_home=None):
    """Return the path of the creme data directory."""
    if data_home is None:
        data_home = os.environ.get('CREME_DATA', os.path.join('~', 'creme_data'))
        data_home = os.path.expanduser(data_home)
        if not os.path.exists(data_home):
            os.makedirs(data_home)
    return data_home


def download_dataset(name, url, data_home, archive_type=None, silent=True):
    """Downloads/decompresses a dataset locally if does not exist.

    Parameters:
        name (str): Dataset filename or directory name.
        url (str): From where to download the dataset.
        data_home (str): The directory where you wish to store the data.
        archive_type (str): Dataset archive type extension name (e.g. 'zip'). Defaults to None.
        silent (bool): Whether to indicate download progress or not.

    Returns:
        data_dir_path (str): Where the dataset is stored.

    """
    data_home = get_data_home(data_home=data_home)
    data_dir_path = os.path.join(data_home, f'{name}')

    # Download dataset if does not exist
    if not os.path.exists(data_dir_path):
        if not silent:
            print(f'Downloading {name}...')

        if archive_type:
            data_dir_path = f'{data_dir_path}.{archive_type}'

        with urllib.request.urlopen(url) as r, open(data_dir_path, 'wb') as f:
            shutil.copyfileobj(r, f)

        # Uncompress if needed
        if archive_type:
            archive_path, data_dir_path = data_dir_path, data_dir_path[:-len(archive_type) - 1]

            if not silent:
                print(f'Uncompressing {name}...')

            if archive_type == 'zip':
                with zipfile.ZipFile(archive_path, 'r') as zf:
                    zf.extractall(data_dir_path)

            elif archive_type in ['gz', 'tar', 'tar.gz', 'tgz']:
                mode = 'r:' if archive_type == 'tar' else 'r:gz'
                tar = tarfile.open(archive_path, mode)
                tar.extractall(data_dir_path)
                tar.close()

            # Delete the archive file now that the dataset is available
            os.remove(archive_path)

    return data_dir_path


def fetch_bikes(data_home=None, silent=True):
    """Bike sharing station information from the city of Toulouse.

    The data contains 182,470 items and 8 features. The goal is to predict the number of bikes in
    5 different bike stations from the city of Toulouse.

    Parameters:
        data_home (str): The directory where you wish to store the data.
        silent (bool): Whether to indicate download progress or not.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `A short introduction and conclusion to the OpenBikes 2016 Challenge <https://maxhalford.github.io/blog/a-short-introduction-and-conclusion-to-the-openbikes-2016-challenge/>`_

    """

    name = 'toulouse_bikes'
    url = 'https://maxhalford.github.io/files/datasets/toulouse_bikes.zip'

    # Download dataset if does not exist and get its path
    data_dir_path = download_dataset(name, url, data_home, archive_type='zip', silent=silent)

    return stream.iter_csv(
        f'{data_dir_path}/{name}.csv',
        target_name='bikes',
        converters={
            'clouds': int,
            'humidity': int,
            'pressure': float,
            'temperature': float,
            'wind': float,
            'bikes': int
        },
        parse_dates={'moment': '%Y-%m-%d %H:%M:%S'}
    )


def fetch_electricity(data_home=None, silent=True):
    """A day of electricity prices in New South Wales.

    This data was collected from the Australian New South Wales Electricity Market. In this market,
    prices are not fixed and are affected by demand and supply of the market. They are set every
    five minutes. Electricity transfers to/from the neighboring state of Victoria were done to
    alleviate fluctuations.

    The dataset (originally named ELEC2) contains 45,312 instances dated from 7 May 1996 to 5
    December 1998. Each example of the dataset refers to a period of 30 minutes, i.e. there are 48
    instances for each time period of one day. Each example on the dataset has 5 fields, the day of
    week, the time stamp, the New South Wales electricity demand, the Victoria electricity demand,
    the scheduled electricity transfer between states and the class label. The class label
    identifies the change of the price (UP or DOWN) in New South Wales relative to a moving average
    of the last 24 hours (and removes the impact of longer term price trends).

    Parameters:
        data_home (str): The directory where you wish to store the data.
        silent (bool): Whether to indicate download progress or not.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `SPLICE-2 Comparative Evaluation: Electricity Pricing <http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.12.9405>`_
        2. `DataHub description <https://datahub.io/machine-learning/electricity#readme>`_

    """

    name = 'electricity'
    url = 'https://maxhalford.github.io/files/datasets/electricity.zip'

    # Download dataset if does not exist and get its path
    data_dir_path = download_dataset(name, url, data_home, archive_type='zip', silent=silent)

    return stream.iter_csv(
        f'{data_dir_path}/{name}.csv',
        target_name='class',
        converters={
            'date': float,
            'day': int,
            'period': float,
            'nswprice': float,
            'nswdemand': float,
            'vicprice': float,
            'vicdemand': float,
            'transfer': float,
            'class': lambda x: x == 'UP'
        }
    )


def fetch_kdd99_http(data_home=None, silent=True):
    """Data from the HTTP dataset of the KDD'99 cup.

    The data contains 567,498 items and 3 features. The goal is to predict whether or not an HTTP
    connection is anomalous or not. The dataset only contains 2,211 (0.4%) positive labels.

    Parameters:
        data_home (str): The directory where you wish to store the data.
        silent (bool): Whether to indicate download progress or not.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `HTTP (KDDCUP99) dataset <http://odds.cs.stonybrook.edu/http-kddcup99-dataset/>`_

    """

    name = 'kdd99_http'
    url = 'https://maxhalford.github.io/files/datasets/kdd99_http.zip'

    # Download dataset if does not exist and get its path
    data_dir_path = download_dataset(name, url, data_home, archive_type='zip', silent=silent)

    return stream.iter_csv(
        f'{data_dir_path}/{name}.csv',
        target_name='service',
        converters={
            'duration': float,
            'src_bytes': float,
            'dst_bytes': float,
            'service': int
        }
    )


def fetch_restaurants(data_home=None, silent=True):
    """Data from the Kaggle Recruit Restaurants challenge.

    The data contains 252,108 items and 7 features. The goal is to predict the number of visitors
    in each of 829 Japanese restaurants over a priod of roughly 16 weeks. The data is ordered by
    date and then by restaurant ID.

    Parameters:
        data_home (str): The directory where you wish to store the data.
        silent (bool): Whether to indicate download progress or not.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `Recruit Restaurant Visitor Forecasting <https://www.kaggle.com/c/recruit-restaurant-visitor-forecasting>`_

    """

    name = 'kaggle_recruit_restaurants'
    url = 'https://maxhalford.github.io/files/datasets/kaggle_recruit_restaurants.zip'

    # Download dataset if does not exist and get its path
    data_dir_path = download_dataset(name, url, data_home, archive_type='zip', silent=silent)

    return stream.iter_csv(
        f'{data_dir_path}/{name}.csv',
        target_name='visitors',
        converters={
            'latitude': float,
            'longitude': float,
            'visitors': int,
            'is_holiday': ast.literal_eval
        },
        parse_dates={'date': '%Y-%m-%d'}
    )


def fetch_sms(data_home=None, silent=True):
    """SMS Spam Collection dataset.

    The data contains 5,574 items and 1 feature (i.e. SMS body). Spam messages represent
    13.4% of the dataset. The goal is to predict whether an SMS is a spam or not.

    Parameters:
        data_home (str): The directory where you wish to store the data.
        silent (bool): Whether to indicate download progress or not.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `Contributions to the Study of SMS Spam Filtering: New Collection and Results <http://www.dt.fee.unicamp.br/~tiago/smsspamcollection/doceng11.pdf>`_

    """

    name = 'SMSSpamCollection'
    url = 'https://archive.ics.uci.edu/ml/machine-learning-databases/00228/smsspamcollection.zip'

    # Download dataset if does not exist and get its path
    data_dir_path = download_dataset(name, url, data_home, archive_type='zip', silent=silent)

    # Stream SMS
    with open(f'{data_dir_path}/{name}') as f:
        for ix, row in enumerate(f):
            label, body = row.split('\t')
            yield ({'body': body}, label == 'spam')


def fetch_trec07p(data_home=None, silent=True):
    """2007 TREC’s Spam Track dataset.

    The data contains 75,419 chronologically ordered items, i.e. 3 months of emails delivered
    to a particular server in 2007. Spam messages represent 66.6% of the dataset.
    The goal is to predict whether an email is a spam or not.

    The available raw features are: sender, recipients, date, subject, body.

    Parameters:
        data_home (str): The directory where you wish to store the data.
        silent (bool): Whether to indicate download progress or not.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `TREC 2007 Spam Track Overview <https://trec.nist.gov/pubs/trec16/papers/SPAM.OVERVIEW16.pdf>`_
        2. `Code ran to parse the dataset <https://gist.github.com/gbolmier/b6a942699aaaedec54041a32e4f34d40>`_

    """

    name = 'trec07p'
    url = 'https://archive.org/download/trec07p/trec07p.tgz'

    # Download dataset if does not exist and get its path
    data_dir_path = download_dataset(name, url, data_home, archive_type='zip', silent=silent)

    return stream.iter_csv(f'{data_dir_path}/{name}.csv', target_name='y')


def fetch_credit_card(data_home=None, silent=True):
    """Credit card fraud dataset.

    The datasets contains transactions made by credit cards in September 2013 by european
    cardholders. This dataset presents transactions that occurred in two days, where we have 492
    frauds out of 284,807 transactions. The dataset is highly unbalanced, the positive class
    (frauds) account for 0.172% of all transactions.

    It contains only numerical input variables which are the result of a PCA transformation.
    Unfortunately, due to confidentiality issues, we cannot provide the original features and more
    background information about the data. Features V1, V2, ... V28 are the principal components
    obtained with PCA, the only features which have not been transformed with PCA are 'Time' and
    'Amount'. Feature 'Time' contains the seconds elapsed between each transaction and the first
    transaction in the dataset. The feature 'Amount' is the transaction Amount, this feature can be
    used for example-dependant cost-senstive learning. Feature 'Class' is the response variable and
    it takes value 1 in case of fraud and 0 otherwise.

    Parameters:
        data_home (str): The directory where you wish to store the data.
        silent (bool): Whether to indicate download progress or not.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. Andrea Dal Pozzolo, Olivier Caelen, Reid A. Johnson and Gianluca Bontempi. Calibrating Probability with Undersampling for Unbalanced Classification. In Symposium on Computational Intelligence and Data Mining (CIDM), IEEE, 2015
        2. Dal Pozzolo, Andrea; Caelen, Olivier; Le Borgne, Yann-Ael; Waterschoot, Serge; Bontempi, Gianluca. Learned lessons in credit card fraud detection from a practitioner perspective, Expert systems with applications,41,10,4915-4928,2014, Pergamon
        3. Dal Pozzolo, Andrea; Boracchi, Giacomo; Caelen, Olivier; Alippi, Cesare; Bontempi, Gianluca. Credit card fraud detection: a realistic modeling and a novel learning strategy, IEEE transactions on neural networks and learning systems,29,8,3784-3797,2018,IEEE
        4. Dal Pozzolo, Andrea Adaptive Machine learning for credit card fraud detection ULB MLG PhD thesis (supervised by G. Bontempi)
        5. Carcillo, Fabrizio; Dal Pozzolo, Andrea; Le Borgne, Yann-Ael; Caelen, Olivier; Mazzer, Yannis; Bontempi, Gianluca. Scarff: a scalable framework for streaming credit card fraud detection with Spark, Information fusion,41, 182-194,2018,Elsevier
        6. Carcillo, Fabrizio; Le Borgne, Yann-Ael; Caelen, Olivier; Bontempi, Gianluca. Streaming active learning strategies for real-life credit card fraud detection: assessment and visualization, International Journal of Data Science and Analytics, 5,4,285-300,2018,Springer International Publishing
        7. Bertrand Lebichot, Yann-Ael Le Borgne, Liyun He, Frederic Oble, Gianluca Bontempi Deep-Learning Domain Adaptation Techniques for Credit Cards Fraud Detection, INNSBDDL 2019: Recent Advances in Big Data and Deep Learning, pp 78-88, 2019
        8. Fabrizio Carcillo, Yann-Ael Le Borgne, Olivier Caelen, Frederic Oble, Gianluca Bontempi Combining Unsupervised and Supervised Learning in Credit Card Fraud Detection Information Sciences, 2019

    """

    name = 'creditcard'
    url = 'https://maxhalford.github.io/files/datasets/creditcardfraud.zip'

    data_dir_path = download_dataset(name, url, data_home, archive_type='zip', silent=silent)

    converters = {f'V{i}': float for i in range(1, 29)}
    converters['Class'] = int
    converters['Time'] = int
    converters['Amount'] = float

    return stream.iter_csv(
        f'{data_dir_path}/{name}.csv',
        target_name='Class',
        converters=converters
    )


def load_airline():
    """Monthly number of international airline passengers.

    The stream contains 144 items and only one single feature, which is the month. The goal is to
    predict the number of passengers each month by capturing the trend and the seasonality of the
    data.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `International airline passengers: monthly totals in thousands. Jan 49 – Dec 60 <https://datamarket.com/data/set/22u3/international-airline-passengers-monthly-totals-in-thousands-jan-49-dec-60#!ds=22u3&display=line>`_

    """
    return stream.iter_csv(
        os.path.join(os.path.dirname(__file__), 'airline-passengers.csv'),
        target_name='passengers',
        converters={'passengers': int},
        parse_dates={'month': '%Y-%m'}
    )


def load_chick_weights():
    """Chick weights along time.

    The stream contains 578 items and 3 features. The goal is to predict the weight of each chick
    along time, according to the diet the chick is on. The data is ordered by time and then by
    chick.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `Chick Weight <http://rstudio-pubs-static.s3.amazonaws.com/107631_131ad1c022df4f90aa2d214a5c5609b2.html>`_

    """
    return stream.iter_csv(
        os.path.join(os.path.dirname(__file__), 'chick-weights.csv'),
        target_name='weight',
        converters={'time': int, 'weight': int, 'chick': int, 'diet': int}
    )



try:
    import surprise
    SURPRISE_INSTALLED = True
except ImportError:
    SURPRISE_INSTALLED = False


if SURPRISE_INSTALLED:

    class SurpriseWrapper(surprise.AlgoBase):

        def __init__(self, creme_recommender):
            self.creme_recommender = creme_recommender

        def fit(self, trainset):
            """Fits to an entire training set.

            This method is compatible with the API from the `surprise` library.

            """
            surprise.AlgoBase.fit(self, trainset)
            for triplet in trainset.all_ratings():
                self.creme_recommender.fit_one(*triplet)
            return self

        def estimate(self, *_):
            return self.creme_recommender.predict_one(*_)



from scipy import integrate

from . import base
from . import confusion


__all__ = ['ROCAUC']


class ROCAUC(base.BinaryMetric):
    """Receiving Operating Characteristic Area Under the Curve.

    Parameters:
        num_thresholds (int): The number of thresholds to use to discretize the ROC curve. The
            higher this is, the closer the output will be to the true ROC AUC value, at the cost of
            more time and memory.

    Attributes:
        thresholds (list)
        cms (list): Contains the `metrics.ConfusionMatrix` for each threshold.

    .. warning::
        This metric is an approximation of the true ROC AUC. Computing the true ROC AUC would
        require storing all the predictions and true outputs, which isn't an option. The
        approximation error is not significant as long as the predicted probabilities are well
        calibrated.

    """

    def __init__(self, num_thresholds=10):
        self.num_thresholds = num_thresholds
        self.thresholds = [i / (num_thresholds - 1) for i in range(num_thresholds)]
        self.cms = [confusion.ConfusionMatrix() for _ in range(num_thresholds)]

    def update(self, y_true, y_pred, sample_weight=1.):
        for t, cm in zip(self.thresholds, self.cms):
            cm.update(y_true=y_true, y_pred=y_pred.get(True, 0.) > t, sample_weight=1.)
        return self

    def revert(self, y_true, y_pred):
        for t, cm in zip(self.thresholds, self.cms):
            cm.update(y_true=y_true, y_pred=y_pred.get(True, 0.) > t, sample_weight=1.)
        return self

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return False

    def get(self):

        tprs = [0] * self.num_thresholds
        fprs = [0] * self.num_thresholds

        def safe_div(a, b):
            try:
                return a / b
            except ZeroDivisionError:
                return 0.

        for i, cm in enumerate(reversed(self.cms)):
            tp = cm.counts.get(True, {}).get(True, 0)
            tn = cm.counts.get(False, {}).get(False, 0)
            fp = cm.counts.get(False, {}).get(True, 0)
            fn = cm.counts.get(True, {}).get(False, 0)

            tprs[i] = safe_div(tp, tp + fn)
            fprs[i] = safe_div(fp, fp + tn)

        # Remove duplicate values
        tprs, fprs = zip(*[(tprs[0], fprs[0])] + [
            (tprs[i + 1], b)
            for i, (a, b) in enumerate(zip(fprs[:-1], fprs[1:]))
            if a != b
        ])

        return integrate.simps(x=fprs, y=tprs)



import collections

from .. import base
from .. import stats
from .. import utils


__all__ = ['MinMaxScaler', 'Normalizer', 'StandardScaler']


def safe_div(a, b):
    if b == 0:
        return a
    return a / b


class StandardScaler(base.Transformer):
    """Scales the data so that it has zero mean and unit variance.

    Under the hood a running mean and a running variance are maintained. The scaling is slightly
    different than when using scikit-learn but this doesn't seem to have any impact on learning
    performance.

    Attributes:
        variances (dict): Mapping between features and instances of `stats.Var`.

    Example:

        ::

              >>> import creme
              >>> import numpy as np
              >>> from sklearn import preprocessing

              >>> rng = np.random.RandomState(42)
              >>> X = [{'x': v} for v in rng.uniform(low=8, high=12, size=15)]

              >>> scaler = creme.preprocessing.StandardScaler()
              >>> for x in X:
              ...     print(scaler.fit_one(x).transform_one(x))
              {'x': 0.0}
              {'x': 0.7071067811865474}
              {'x': 0.15899361505958234}
              {'x': -0.27053221516496234}
              {'x': -1.3161741269826}
              {'x': -1.051218830939811}
              {'x': -1.1096762396286515}
              {'x': 1.09141268517007}
              {'x': 0.31090608502985784}
              {'x': 0.5949866915752465}
              {'x': -1.3540960661327464}
              {'x': 1.2959176347038681}
              {'x': 0.8426620966002304}
              {'x': -0.8843412116857523}
              {'x': -0.91188180555936}

              >>> X = np.array([x['x'] for x in X]).reshape(-1, 1)
              >>> preprocessing.StandardScaler().fit_transform(X)
              array([[-0.36224883],
                     [ 1.37671717],
                     [ 0.71659166],
                     [ 0.31416852],
                     [-1.02177407],
                     [-1.02184687],
                     [-1.31735428],
                     [ 1.1215704 ],
                     [ 0.32158263],
                     [ 0.64439399],
                     [-1.43053132],
                     [ 1.43465174],
                     [ 1.01975844],
                     [-0.85179183],
                     [-0.94388734]])

    """

    def __init__(self):
        self.variances = collections.defaultdict(stats.Var)

    def fit_one(self, x, y=None):

        for i, xi in x.items():
            self.variances[i].update(xi)

        return self

    def transform_one(self, x):
        return {
            i: safe_div(xi - self.variances[i].mean.get(), self.variances[i].get() ** .5)
            for i, xi in x.items()
        }


class MinMaxScaler(base.Transformer):
    """Scales the data to a fixed range from 0 to 1.

    Under the hood a running min and a running peak to peak (max - min) are maintained. The scaling
    is slightly different than when using scikit-learn but this doesn't seem to have any impact on
    learning performance.

    Attributes:
        min (dict): Mapping between features and instances of `stats.Min`.
        max (dict): Mapping between features and instances of `stats.Max`.

    Example:

        ::

            >>> import creme
            >>> import numpy as np
            >>> from sklearn import preprocessing

            >>> rng = np.random.RandomState(42)
            >>> X = [{'x': v} for v in rng.uniform(low=8, high=12, size=15)]

            >>> scaler = creme.preprocessing.MinMaxScaler()
            >>> for x in X:
            ...     print(scaler.fit_one(x).transform_one(x))
            {'x': 0.0}
            {'x': 1.0}
            {'x': 0.6203919416734277}
            {'x': 0.3889767542308411}
            {'x': 0.0}
            {'x': 0.0}
            {'x': 0.0}
            {'x': 0.9052932403284701}
            {'x': 0.6083494586037179}
            {'x': 0.7281723223510779}
            {'x': 0.0}
            {'x': 1.0}
            {'x': 0.8551948389216542}
            {'x': 0.20199040802352813}
            {'x': 0.16984743067826635}

            >>> X = np.array([x['x'] for x in X]).reshape(-1, 1)
            >>> preprocessing.MinMaxScaler().fit_transform(X)
            array([[0.37284965],
                   [0.9797798 ],
                   [0.74938422],
                   [0.60893137],
                   [0.14266357],
                   [0.14263816],
                   [0.03950081],
                   [0.89072903],
                   [0.61151903],
                   [0.72418595],
                   [0.        ],
                   [1.        ],
                   [0.85519484],
                   [0.20199041],
                   [0.16984743]])

    """

    def __init__(self):
        self.min = collections.defaultdict(stats.Min)
        self.max = collections.defaultdict(stats.Max)

    def fit_one(self, x, y=None):

        for i, xi in x.items():
            self.min[i].update(xi)
            self.max[i].update(xi)

        return self

    def transform_one(self, x):
        return {
            i: safe_div(xi - self.min[i].get(), self.max[i].get() - self.min[i].get())
            for i, xi in x.items()
        }


class Normalizer(base.Transformer):
    """Scales a set of features so that it has unit norm.

    This is particularly useful when used after a `feature_extraction.TFIDFVectorizer`.

    Parameters:
        order (int): Order of the norm (e.g. 2 corresponds to the $L^2$ norm).

    Example:

        ::

            >>> from creme import preprocessing
            >>> from creme import stream

            >>> scaler = preprocessing.Normalizer(order=2)

            >>> X = [[4, 1, 2, 2],
            ...      [1, 3, 9, 3],
            ...      [5, 7, 5, 1]]

            >>> for x, _ in stream.iter_array(X):
            ...     print(scaler.transform_one(x))
            {0: 0.8, 1: 0.2, 2: 0.4, 3: 0.4}
            {0: 0.1, 1: 0.3, 2: 0.9, 3: 0.3}
            {0: 0.5, 1: 0.7, 2: 0.5, 3: 0.1}

    """

    def __init__(self, order=2):
        self.order = order

    def transform_one(self, x):
        norm = utils.math.norm(x, order=self.order)
        return {i: xi / norm for i, xi in x.items()}



from sklearn import utils

from .. import stats

from . import base


__all__ = ['RandomNormal']


class RandomNormal(base.Recommender):
    """Predicts random values sampled from a normal distribution.

    The parameters of the normal distribution are fitted with running statistics. This is
    equivalent to using `surprise.prediction_algorithms.random_pred.NormalPredictor`.

    Parameters:
        random_state (int, ``numpy.random.RandomState`` instance or None): If int, ``random_state``
            is the seed used by the random number generator; if ``RandomState`` instance,
            ``random_state`` is the random number generator; if ``None``, the random number
            generator is the ``RandomState`` instance used by `numpy.random`.

    Attributes:
        variance (stats.Variance)

    """

    def __init__(self, random_state=None):
        super().__init__()
        self.variance = stats.Variance()
        self.random_state = utils.check_random_state(random_state)

    def fit_one(self, r_id, c_id, y):
        y_pred = self.predict_one(r_id, c_id)
        self.variance.update(y)
        return y_pred

    def predict_one(self, r_id, c_id):
        μ = self.variance.mean.get() or 0
        σ = self.variance.get() ** 0.5
        return self.random_state.normal(μ, σ)



import bisect
import collections
import math


__all__ = ['Histogram']


class Bin:
    """A Bin is an element of a Histogram."""

    __slots__ = ['left', 'right', 'count']

    def __init__(self, left, right=None, count=1):
        self.left = left
        self.right = left if right is None else right
        self.count = count

    def __add__(self, other):
        return Bin(
            left=min(self.left, other.left),
            right=max(self.right, other.right),
            count=self.count + other.count
        )

    def __lt__(self, other):
        return self.right < other.left

    def __eq__(self, other):
        return self.left == other.left and self.right == other.right

    def __repr__(self):
        return f'[{self.left:.5f}, {self.right:.5f}]: {self.count}'


class Histogram(collections.UserList):
    """Streaming histogram.

    Parameters:
        max_bins (int): Maximal number of bins.

    Attributes:
        n (int): Total number of seen values.

    Example:

        ::

            >>> from creme import utils
            >>> import matplotlib.pyplot as plt
            >>> import numpy as np

            >>> np.random.seed(42)

            >>> values = np.hstack((
            ...     np.random.normal(-3, 1, 1000),
            ...     np.random.normal(3, 1, 1000),
            ... ))

            >>> hist = utils.Histogram(max_bins=60)

            >>> for x in values:
            ...     hist = hist.update(x)

            >>> ax = plt.bar(
            ...     x=[(b.left + b.right) / 2 for b in hist],
            ...     height=[b.count for b in hist],
            ...     width=[(b.right - b.left) / 2 for b in hist]
            ... )

        .. image:: ../_static/histogram_docstring.svg
            :align: center

    References:
        1. `A Streaming Parallel Decision Tree Algorithm <http://jmlr.org/papers/volume11/ben-haim10a/ben-haim10a.pdf>`_
        2. `Go implementation <https://github.com/VividCortex/gohistogram>`_

    """

    def __init__(self, max_bins=256):
        super().__init__()
        self.max_bins = max_bins
        self.n = 0

    def update(self, x):

        self.n += 1
        b = Bin(x)

        # Insert the bin if the histogram is empty
        if not self:
            self.append(b)
            return self

        # Use bisection to find where to insert
        lo = 0
        hi = len(self)
        i = (lo + hi) // 2
        while lo < hi:
            if self[i] < b:
                lo = i + 1
            else:
                hi = i
            i = (lo + hi) // 2

        if i == len(self):
            # x is past the right-most bin
            self.append(b)
        else:
            # Increment the bin counter if x is part of the ith bin
            if x >= self[i].left:
                self[i].count += 1
            # Insert the bin if it is between bin i-1 and bin i
            else:
                self.insert(i, b)

        # Bins have to be merged if there are more than max_bins
        if len(self) > self.max_bins:

            # Find the closest pair of bins
            min_diff = math.inf
            min_idx = None
            for idx, (b1, b2) in enumerate(zip(self.data[:-1], self.data[1:])):
                diff = b2.right - b1.right
                if diff < min_diff:
                    min_diff = diff
                    min_idx = idx

            # Merge the bins
            self[min_idx] += self.pop(min_idx + 1)

        return self

    def cdf(self, x):
        """Cumulative distribution function.

        Example:

            >>> from creme import utils

            >>> hist = Histogram()
            >>> for x in range(4):
            ...     hist = hist.update(x)

            >>> print(hist)
            [0.00000, 0.00000]: 1
            [1.00000, 1.00000]: 1
            [2.00000, 2.00000]: 1
            [3.00000, 3.00000]: 1

            >>> hist.cdf(0)
            0.25

            >>> hist.cdf(.5)
            0.375

            >>> hist.cdf(1)
            0.5

            >>> hist.cdf(2.5)
            0.875

        """

        # Handle edge cases
        if not self or x < self[0].left:
            return 0.
        elif x >= self[-1].right:
            return 1.

        c = 0

        # Handle the first bin
        b = self[0]
        if x < b.right:
            c += b.count * (x - b.left) / (b.right - b.left)
            return c / self.n
        c += b.count

        # Handle the rest of the bins
        for b1, b2 in zip(self.data[:-1], self.data[1:]):
            if x < b2.right:
                # Interpolate
                c += b2.count * (x - b1.right) / (b2.right - b1.right)
                break
            c += b2.count

        return c / self.n

    def __repr__(self):
        return '\n'.join(str(b) for b in self)



from .. import stats

from . import base


__all__ = ['MAE']


class MAE(stats.Mean, base.RegressionMetric):
    """Mean absolute error.

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = [3, -0.5, 2, 7]
            >>> y_pred = [2.5, 0.0, 2, 8]

            >>> metric = metrics.MAE()

            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     print(metric.update(y_t, y_p).get())
            0.5
            0.5
            0.333333...
            0.5

            >>> metric
            MAE: 0.5

    """

    def update(self, y_true, y_pred, sample_weight=1.):
        return super().update(x=abs(y_true - y_pred), w=sample_weight)

    def revert(self, y_true, y_pred, sample_weight=1.):
        return super().revert(x=abs(y_true - y_pred), w=sample_weight)



import math

from .. import utils

from . import base


class Max(base.Univariate):
    """Running max.

    Attributes:
        max (float): The running max.

    """

    def __init__(self):
        self.max = -math.inf

    def update(self, x):
        if x > self.max:
            self.max = x
        return self

    def get(self):
        return self.max


class RollingMax(base.RollingUnivariate, utils.SortedWindow):
    """Running max over a window.

    Parameters:
        window_size (int): Size of the rolling window.

    Example:

        ::

            >>> from creme import stats

            >>> X = [1, -4, 3, -2, 2, 1]
            >>> rolling_max = stats.RollingMax(window_size=2)
            >>> for x in X:
            ...     print(rolling_max.update(x).get())
            1
            1
            3
            3
            2
            2

    """

    def __init__(self, window_size):
        super().__init__(size=window_size)

    @property
    def window_size(self):
        return self.size

    def update(self, x):
        self.append(x)
        return self

    def get(self):
        return self[-1]
